--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;

-- Purpose:
-- . Keep a running sum for each element in snk_in_arr[i]
-- Description:
-- . A RAM-block is used to store the running sums as wel as to output the
--   corresponding previous running sum so both align at the adder inputs.
-- . g_nof_words_to_acc=0 connects the source to the sink.

ENTITY corr_accumulator IS
  GENERIC (
    g_nof_inputs        : NATURAL;
    g_nof_acc_per_input : NATURAL;
    g_nof_words_to_acc  : NATURAL;
    g_data_w            : NATURAL
   ); 
  PORT (
    rst            : IN  STD_LOGIC;
    clk            : IN  STD_LOGIC;

    snk_in_arr     : IN  t_dp_sosi_arr;
    src_out_arr    : OUT t_dp_sosi_arr
  );
END corr_accumulator;

ARCHITECTURE str OF corr_accumulator IS

  -- Help constant so other constants can be calculated if g_nof_words_to_acc<1
--  CONSTANT g_nof_words_to_acc        : NATURAL := sel_a_b(g_nof_words_to_acc>0, g_nof_words_to_acc, 3);

  CONSTANT c_shiftram_io_delay        : NATURAL := 4; -- common_shiftram data_in takes 4 cycles to emerge as data_out

  CONSTANT c_shiftram_data_w          : NATURAL := 64; --2*(g_data_w+1);

  -- c_shiftram_delay is such that common_shiftram output aligns exactly with snk_in_arr. Functionally this 
  -- means we aligned the current word to the corresponding previous word at the adder inputs.
  CONSTANT c_shiftram_delay           : NATURAL := g_nof_acc_per_input-c_shiftram_io_delay-1;
  CONSTANT c_shift_w                  : NATURAL := ceil_log2(g_nof_acc_per_input);
  CONSTANT c_common_shiftram_shift_in : STD_LOGIC_VECTOR(c_shift_w-1 DOWNTO 0) := TO_UVEC(c_shiftram_delay, c_shift_w);

  SIGNAL corr_adder_snk_in_2arr_2     : t_dp_sosi_2arr_2(g_nof_inputs-1 DOWNTO 0); -- Array of pairs
  SIGNAL corr_adder_src_out_arr       : t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);

  SIGNAL common_shiftram_snk_in_arr   : t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);
  SIGNAL common_shiftram_src_out_arr  : t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);

  TYPE t_common_shiftram_data_out_shift_arr IS ARRAY(g_nof_inputs-1 DOWNTO 0) OF STD_LOGIC_VECTOR(c_shift_w-1 DOWNTO 0);
  SIGNAL common_shiftram_data_out_shift_arr : t_common_shiftram_data_out_shift_arr;

BEGIN

--  gen_accumulator: IF g_nof_words_to_acc>0 GENERATE
    -----------------------------------------------------------------------------
    -- Adder inputs: current snk_in_arr + corresponding previous running sum
    -- from shiftram
    -----------------------------------------------------------------------------
    gen_adder_inputs : FOR i IN 0 TO g_nof_inputs-1 GENERATE
      corr_adder_snk_in_2arr_2(i)(0) <= snk_in_arr(i);
      corr_adder_snk_in_2arr_2(i)(1).valid <= common_shiftram_src_out_arr(i).valid;
      corr_adder_snk_in_2arr_2(i)(1).re(31-1 DOWNTO 0) <= common_shiftram_src_out_arr(i).data(c_shiftram_data_w-1-1   DOWNTO c_shiftram_data_w/2);
      corr_adder_snk_in_2arr_2(i)(1).im(31-1 DOWNTO 0) <= common_shiftram_src_out_arr(i).data(c_shiftram_data_w/2-1-1 DOWNTO 0);
    END GENERATE;
  
    -----------------------------------------------------------------------------
    -- Complex adder stage
    -----------------------------------------------------------------------------
    u_corr_adder : ENTITY work.corr_adder
    GENERIC MAP (
      g_nof_inputs => g_nof_inputs,
      g_data_w     => 64/2-1 --So output becomes 32 bit wide (64 concatenated)
    )
    PORT MAP (
      clk            => clk,
      rst            => rst,
  
      snk_in_2arr_2  => corr_adder_snk_in_2arr_2,
      src_out_arr    => corr_adder_src_out_arr
    );  
  
    -----------------------------------------------------------------------------
    -- Write the current sum to RAM; RAM outputs delayed running sums that align
    -- at the adder inputs:
    -- . common_shiftram_src_out_arr = delayed corr_adder_src_out_arr
    -----------------------------------------------------------------------------
    gen_concat_complex : FOR i IN 0 TO g_nof_inputs-1 GENERATE
      -- Concatenate real&imaginary parts
      common_shiftram_snk_in_arr(i).data(c_shiftram_data_w-1   DOWNTO c_shiftram_data_w/2) <= corr_adder_src_out_arr(i).re(c_shiftram_data_w/2-1 DOWNTO 0);
      common_shiftram_snk_in_arr(i).data(c_shiftram_data_w/2-1 DOWNTO 0)                   <= corr_adder_src_out_arr(i).im(c_shiftram_data_w/2-1 DOWNTO 0);
      common_shiftram_snk_in_arr(i).valid                                                  <= corr_adder_src_out_arr(i).valid;
    END GENERATE;
  
    gen_common_shiftram : FOR i IN 0 TO g_nof_inputs-1 GENERATE
      u_common_shiftram : ENTITY common_lib.common_shiftram
      GENERIC MAP (
        g_data_w    => c_shiftram_data_w,
        g_nof_words => pow2(ceil_log2(g_nof_acc_per_input)),
        g_output_invalid_during_shift_incr => TRUE
      )
      PORT MAP (
        rst            => rst,
        clk            => clk,
        
        data_in        => common_shiftram_snk_in_arr(i).data(c_shiftram_data_w-1 DOWNTO 0),
        data_in_val    => common_shiftram_snk_in_arr(i).valid,
        data_in_shift  => c_common_shiftram_shift_in,
     
        data_out       => common_shiftram_src_out_arr(i).data(c_shiftram_data_w-1 DOWNTO 0),
        data_out_val   => common_shiftram_src_out_arr(i).valid,
        data_out_shift => common_shiftram_data_out_shift_arr(i)
      );
    END GENERATE;
  
    -----------------------------------------------------------------------------
    -- Output 1/g_nof_acc_per_input words per stream
    -- . Not implemented yet.
    -----------------------------------------------------------------------------
    gen_src_out_arr : FOR i IN 0 TO g_nof_inputs-1 GENERATE
      src_out_arr(i).re(c_shiftram_data_w/2-1 DOWNTO 0) <= common_shiftram_src_out_arr(i).data(c_shiftram_data_w-1   DOWNTO c_shiftram_data_w/2);
      src_out_arr(i).im(c_shiftram_data_w/2-1 DOWNTO 0) <= common_shiftram_src_out_arr(i).data(c_shiftram_data_w/2-1 DOWNTO 0);
      src_out_arr(i).valid <= common_shiftram_src_out_arr(i).valid;
    END GENERATE;

--  END GENERATE;

--  gen_bypass: IF g_nof_words_to_acc<1 GENERATE
--    src_out_arr <= snk_in_arr;
--  END GENERATE;

END str;
