-------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_network_layers_pkg.ALL;
USE common_lib.common_network_total_header_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE work.eth_pkg.ALL;

-- Description:
-- . for IP frames replace IP header checksum with the calculated value
-- . discard frames shorter than c_network_total_header_32b_nof_words = 11 words
-- . frame discard input to optionally discard a frame based on some header criteria

ENTITY eth_hdr_ctrl IS
  PORT (
    -- Clocks and reset
    rst             : IN  STD_LOGIC;
    clk             : IN  STD_LOGIC;
    
    -- Stored header
    hdr_words_arr   : IN  t_network_total_header_32b_arr;
    hdr_status      : IN  t_eth_hdr_status;
    
    frm_discard     : IN  STD_LOGIC := '0';
    
    -- ST interface
    snk_in_word_cnt : IN  NATURAL RANGE 0 TO c_network_total_header_32b_nof_words;  -- snk_in hdr word count
    snk_in          : IN  t_dp_sosi;
    snk_out         : OUT t_dp_siso;
    src_in          : IN  t_dp_siso;
    src_out         : OUT t_dp_sosi
  );
END eth_hdr_ctrl;


ARCHITECTURE rtl OF eth_hdr_ctrl IS

  -- Internal sink ready latency and source ready latency of this component
  CONSTANT c_this_snk_latency : NATURAL := 1;
  CONSTANT c_this_src_latency : NATURAL := 1;
    
  TYPE t_state_enum IS (
    s_idle,
    s_store_header,
    s_src_header_sop,
    s_src_header,
    s_src_data
  );
  
  SIGNAL state               : t_state_enum;
  SIGNAL nxt_state           : t_state_enum;
  
  -- Sink
  SIGNAL snk_in_eop_hold     : STD_LOGIC;
  SIGNAL nxt_snk_in_eop_hold : STD_LOGIC;
  
  -- Source
  SIGNAL next_src_out        : t_dp_sosi;
  SIGNAL i_src_out           : t_dp_sosi;
  SIGNAL nxt_src_out         : t_dp_sosi;
  SIGNAL src_word_cnt        : NATURAL RANGE 0 TO c_network_total_header_32b_nof_words;
  SIGNAL nxt_src_word_cnt    : NATURAL;
  
BEGIN

  src_out <= i_src_out;
  
  p_reg : PROCESS(rst, clk)
  BEGIN
    IF rst='1' THEN
      state           <= s_idle;
      snk_in_eop_hold <= '0';
      src_word_cnt    <= 0;
      i_src_out       <= c_dp_sosi_rst;
    ELSIF rising_edge(clk) THEN
      state           <= nxt_state;
      snk_in_eop_hold <= nxt_snk_in_eop_hold;
      src_word_cnt    <= nxt_src_word_cnt;
      i_src_out       <= nxt_src_out;
    END IF;
  END PROCESS;
  
  
  ------------------------------------------------------------------------------
  -- Hold the sink input for source output
  ------------------------------------------------------------------------------
  
  u_snk_in : ENTITY dp_lib.dp_hold_input
  PORT MAP (
    rst           => rst,
    clk           => clk,
    -- ST sink
    snk_out       => OPEN,
    snk_in        => snk_in,
    -- ST source
    src_in        => src_in,  -- must not use snk_out.ready here to avoid wrong first payload data
    next_src_out  => next_src_out,
    pend_src_out  => OPEN,
    src_out_reg   => i_src_out
  );  
    
  ------------------------------------------------------------------------------
  -- Control state machine for the source output
  ------------------------------------------------------------------------------
  
  -- Hold snk_in.eop, because snk_in packet can only have a header of exactly 11 words (e.g. ARP).
  -- Discard snk_in packets that are shorter than 11 words (e.g. ???), for both Rx and Tx shorter frames are not useful.
  nxt_snk_in_eop_hold <= '0' WHEN snk_in.sop='1' ELSE
                         '1' WHEN snk_in.eop='1' ELSE snk_in_eop_hold;
  
  p_state : PROCESS(state, src_word_cnt, i_src_out, snk_in, snk_in_word_cnt, frm_discard, src_in, hdr_words_arr, hdr_status, snk_in_eop_hold, next_src_out)
  BEGIN
    snk_out.ready     <= '1';           -- default accept sink input
    nxt_state         <= state;
    nxt_src_word_cnt  <= src_word_cnt;
    nxt_src_out       <= i_src_out;
    nxt_src_out.valid <= '0';
    nxt_src_out.sop   <= '0';
    nxt_src_out.eop   <= '0';
    CASE state IS
      WHEN s_idle =>
        -- wait for new frame, then store the frame header and no output yet
        IF snk_in.sop='1' THEN
          nxt_src_word_cnt <= 0;
          nxt_state        <= s_store_header;
        END IF;
      WHEN s_store_header =>
        -- stop the sink when the whole header has arrived or when a eop has arrived
        IF snk_in.valid='1' AND snk_in_word_cnt=c_network_total_header_32b_nof_words-1 THEN
          snk_out.ready <= '0';
          nxt_state <= s_src_header_sop;
        ELSIF snk_in.eop='1' THEN
          -- discard not supported packets shorter than 11 words
          snk_out.ready <= '0';
          nxt_state <= s_idle;
        END IF;
      WHEN s_src_header_sop =>
        -- stop the input and output the (modified) header, first the sop
        snk_out.ready <= '0';
        IF frm_discard='1' THEN
          -- discard this frame because of optional external reasons
          nxt_state <= s_idle;
        ELSIF src_in.ready='1' THEN
          nxt_src_out.data  <= RESIZE_DP_DATA(hdr_words_arr(src_word_cnt));
          nxt_src_out.valid <= '1';
          nxt_src_out.sop   <= '1';
          nxt_src_out.eop   <= '0';
          nxt_src_word_cnt  <= src_word_cnt+1;
          nxt_state <= s_src_header;
        END IF;
      WHEN s_src_header =>
        -- stop the input and output the rest of the (modified) header
        snk_out.ready <= '0';
        IF src_in.ready='1' THEN
          nxt_src_out.data  <= RESIZE_DP_DATA(hdr_words_arr(src_word_cnt));
          nxt_src_out.valid <= '1';
          nxt_src_out.sop   <= '0';
          nxt_src_out.eop   <= '0';
          nxt_src_word_cnt  <= src_word_cnt+1;
          -- Replace IP header checksum with the calculated checksum from hdr_status.ip_checksum:
          -- . for Rx: the verified checksum, so will be 0 when OK or != when the received frame has an IP header error.
          -- . for Tx: the calculated checksum for the Tx IP header.
          IF hdr_status.is_ip='1' AND src_word_cnt=c_network_total_header_32b_ip_header_checksum_wi THEN
            nxt_src_out.data(c_network_ip_header_checksum_w-1 DOWNTO 0) <= hdr_status.ip_checksum;
          END IF;
          IF src_word_cnt=c_network_total_header_32b_nof_words-1 THEN
            IF snk_in_eop_hold='1' THEN
              -- header only packet
              nxt_src_out.eop <= '1';
              nxt_state <= s_idle;
            ELSE
              -- packet with payload
              nxt_state <= s_src_data;
            END IF;
          END IF;
        END IF;
      WHEN OTHERS =>  -- s_src_data
        -- continue the input payload and output it
        snk_out.ready <= src_in.ready;
        nxt_src_out   <= next_src_out;
        IF next_src_out.eop='1' THEN
          nxt_state <= s_idle;
        END IF;
    END CASE;
    
    -- Pass on frame level flow control
    snk_out.xon <= src_in.xon;
  END PROCESS;

END rtl;
