--------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib, mdio_lib, diagnostics_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_interface_layers_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE mdio_lib.mdio_pkg.ALL;
USE mdio_lib.mdio_vitesse_vsc8486_pkg.ALL;

ENTITY mms_tr_xaui IS
  GENERIC(
    g_sim             : BOOLEAN := FALSE;
    g_nof_xaui        : NATURAL := 1; -- Up to 3 (hard XAUI only) supported
    g_mdio            : BOOLEAN := FALSE
--    g_mdio_mm_ctrl    : BOOLEAN := FALSE -- FALSE uses mdio_ctlr to auto execute the initialization 
--                                         -- sequence for the Vitesse chips. TRUE enables MM control.
    );                      
  PORT (   
    -- Transceiver PLL reference clock   
    tr_clk                  : IN  STD_LOGIC;

    -- Calibration & reconfig clock
    cal_rec_clk             : IN  STD_LOGIC;
 
    -- MM clock for register of optional MDIO master
    mm_clk                  : IN  STD_LOGIC := '0';
    mm_rst                  : IN  STD_LOGIC := '0';

    -- Streaming RX interfaces
    rx_clk                  : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
    rx_rst                  : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
    rx_sosi_arr             : OUT t_dp_sosi_arr(g_nof_xaui-1 DOWNTO 0);
    rx_siso_arr             : IN  t_dp_siso_arr(g_nof_xaui-1 DOWNTO 0);

    -- Streaming TX interfaces
    tx_clk                  : IN  STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
    tx_rst                  : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
    tx_sosi_arr             : IN  t_dp_sosi_arr(g_nof_xaui-1 DOWNTO 0);
    tx_siso_arr             : OUT t_dp_siso_arr(g_nof_xaui-1 DOWNTO 0);
   
    --Serial I/O
    xaui_rx                 : IN  t_xaui_arr(g_nof_xaui-1 DOWNTO 0);
    xaui_tx                 : OUT t_xaui_arr(g_nof_xaui-1 DOWNTO 0);   
      
    -- MDIO master = mm slave
    mdio_mosi_arr           : IN  t_mem_mosi_arr(g_nof_xaui-1 DOWNTO 0) := (OTHERS=>c_mem_mosi_rst);
    mdio_miso_arr           : OUT t_mem_miso_arr(g_nof_xaui-1 DOWNTO 0);

    -- MDIO External clock and serial data.
    mdio_rst                : OUT STD_LOGIC;
    mdio_mdc                : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);       
    mdio_mdat_in            : IN  STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0) := (OTHERS=>'0');
    mdio_mdat_oen           : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);

    diagnostics_mosi        : IN  t_mem_mosi := c_mem_mosi_rst;
    diagnostics_miso        : OUT t_mem_miso := c_mem_miso_rst;

    xaui_mosi               : IN  t_mem_mosi := c_mem_mosi_rst;
    xaui_miso               : OUT t_mem_miso := c_mem_miso_rst
    );
END mms_tr_xaui;


ARCHITECTURE wrap OF mms_tr_xaui IS  

  CONSTANT c_nof_select    : NATURAL := 2;
  CONSTANT c_nof_select_w  : NATURAL := ceil_log2(c_nof_select);
  CONSTANT c_sel_user      : NATURAL := 0;
  CONSTANT c_sel_diag      : NATURAL := 1;
 
  TYPE t_select_sosi_2arr IS ARRAY (INTEGER RANGE <>) OF t_dp_sosi_arr(c_nof_select-1 DOWNTO 0);
  TYPE t_select_siso_2arr IS ARRAY (INTEGER RANGE <>) OF t_dp_siso_arr(c_nof_select-1 DOWNTO 0);

  TYPE t_select_arr IS ARRAY (INTEGER RANGE <>) OF STD_LOGIC_VECTOR(c_nof_select_w-1 DOWNTO 0);

  SIGNAL i_tx_rst             :  STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
  SIGNAL i_rx_clk             :  STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
  SIGNAL i_rx_rst             :  STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);

  SIGNAL diagnostics_rx_sosi_arr  : t_dp_sosi_arr(g_nof_xaui-1 DOWNTO 0); 
  SIGNAL diagnostics_rx_siso_arr  : t_dp_siso_arr(g_nof_xaui-1 DOWNTO 0);

  SIGNAL diagnostics_tx_sosi_arr  : t_dp_sosi_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL diagnostics_tx_siso_arr  : t_dp_siso_arr(g_nof_xaui-1 DOWNTO 0);

  SIGNAL diagnostics_tx_en_arr    : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
  SIGNAL diagnostics_rx_en_arr    : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);

  SIGNAL mux_in_siso_2arr         : t_select_siso_2arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL mux_in_sosi_2arr         : t_select_sosi_2arr(g_nof_xaui-1 DOWNTO 0);

  SIGNAL mux_out_sosi_arr         : t_dp_sosi_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL mux_out_siso_arr         : t_dp_siso_arr(g_nof_xaui-1 DOWNTO 0);

  SIGNAL demux_in_siso_arr        : t_dp_siso_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL demux_in_sosi_arr        : t_dp_sosi_arr(g_nof_xaui-1 DOWNTO 0);

  SIGNAL demux_out_sosi_2arr      : t_select_sosi_2arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL demux_out_siso_2arr      : t_select_siso_2arr(g_nof_xaui-1 DOWNTO 0);

  SIGNAL mux_select_arr           : t_select_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL nxt_mux_select_arr       : t_select_arr(g_nof_xaui-1 DOWNTO 0);

  SIGNAL demux_select_arr         : t_select_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL nxt_demux_select_arr     : t_select_arr(g_nof_xaui-1 DOWNTO 0);

BEGIN

  rx_rst <= i_rx_rst;
  rx_clk <= i_rx_clk;
  tx_rst <= i_tx_rst;

  u_tr_xaui: ENTITY work.tr_xaui 
  GENERIC MAP (
    g_sim              => g_sim,
    g_mdio             => g_mdio,
    g_nof_xaui         => g_nof_xaui
  )
  PORT MAP (     
    tr_clk             => tr_clk,

    mm_rst             => mm_rst,  
    mm_clk             => mm_clk,
                        
    --Serial data
    xaui_tx            => xaui_tx, 
    xaui_rx            => xaui_rx,  

    mdio_mosi_arr      => mdio_mosi_arr,
    mdio_miso_arr      => mdio_miso_arr,

    mdio_rst           => mdio_rst,
    mdio_mdc           => mdio_mdc,
    mdio_mdat_in       => mdio_mdat_in,
    mdio_mdat_oen      => mdio_mdat_oen, 
 
    --Parallel data
    rx_rst             => i_rx_rst,   
    rx_clk             => i_rx_clk,  
    rx_sosi_arr        => demux_in_sosi_arr,   
    rx_siso_arr        => demux_in_siso_arr,
    
    tx_rst             => i_tx_rst,    
    tx_clk             => tx_clk,   
    tx_sosi_arr        => mux_out_sosi_arr,   
    tx_siso_arr        => mux_out_siso_arr,

    cal_rec_clk        => cal_rec_clk,

    xaui_mosi          => xaui_mosi,
    xaui_miso          => xaui_miso
    );

  u_mms_diagnostics: ENTITY diagnostics_lib.mms_diagnostics
  GENERIC MAP(
    g_data_w       => c_xgmii_data_w,
    g_nof_streams  => g_nof_xaui,
    g_separate_clk => TRUE
  )                      
  PORT MAP (       
    mm_rst          => mm_rst,  
    mm_clk          => mm_clk,

    src_rst         => i_tx_rst,
    src_clk         => tx_clk,

    snk_rst         => i_rx_rst,
    snk_clk         => i_rx_clk,
    
    mm_mosi         => diagnostics_mosi, 
    mm_miso         => diagnostics_miso,

    src_out_arr     => diagnostics_tx_sosi_arr,
    src_in_arr      => diagnostics_tx_siso_arr,

    snk_out_arr     => diagnostics_rx_siso_arr,
    snk_in_arr      => diagnostics_rx_sosi_arr,

    src_en_out      => diagnostics_tx_en_arr,
    snk_en_out      => diagnostics_rx_en_arr
  );

   gen_select : FOR i IN 0 TO g_nof_xaui-1 GENERATE

    -- 0 = user data, 
    tx_siso_arr(i)                     <= mux_in_siso_2arr(i)(c_sel_user);
    mux_in_sosi_2arr(i)(c_sel_user)    <= tx_sosi_arr(i);
  
    demux_out_siso_2arr(i)(c_sel_user) <= rx_siso_arr(i);
    rx_sosi_arr(i)                     <= demux_out_sosi_2arr(i)(c_sel_user);
  
    -- 1 = internal diagnostics data
    diagnostics_tx_siso_arr(i)         <= mux_in_siso_2arr(i)(c_sel_diag);
    mux_in_sosi_2arr(i)(c_sel_diag)    <= diagnostics_tx_sosi_arr(i);
  
    demux_out_siso_2arr(i)(c_sel_diag) <= diagnostics_rx_siso_arr(i);
    diagnostics_rx_sosi_arr(i)         <= demux_out_sosi_2arr(i)(c_sel_diag);

    -- If user enables internal diagnostics source, the dp_mux automatically forwards that data to the corresponding transmitter
    nxt_mux_select_arr(i) <= slv(diagnostics_tx_en_arr(i));

    -- If user enables internal diagnostics sink, the dp_demux is automatically set to forward the corresponding receiver data
    nxt_demux_select_arr(i) <= slv(diagnostics_rx_en_arr(i));

    u_dp_mux : ENTITY dp_lib.dp_mux
    GENERIC MAP (
      -- Mux
      g_sel_ctrl_invert   => TRUE, -- We're using DOWNTO ranges in out streaming arrays, so we must invert the selection input
      g_mode              => 2,  
      g_nof_input         => c_nof_select, 
      g_append_channel_lo => FALSE, 
      -- Input FIFO
      g_use_fifo          => FALSE,
      g_fifo_size         => array_init(1024, c_nof_select),
      g_fifo_fill         => array_init(   0, c_nof_select)
    )
    PORT MAP (
      rst         => i_tx_rst(i),
      clk         => tx_clk(i),
  
      sel_ctrl    => TO_UINT(mux_select_arr(i)),

      -- ST sinks
      snk_out_arr => mux_in_siso_2arr(i), 
      snk_in_arr  => mux_in_sosi_2arr(i),
      -- ST source
      src_in      => mux_out_siso_arr(i),  
      src_out     => mux_out_sosi_arr(i)
    );  

    p_tx_clk : PROCESS(i_tx_rst, tx_clk)
    BEGIN
      IF i_tx_rst(i)='1' THEN
        mux_select_arr(i) <= (OTHERS=>'0');
      ELSIF rising_edge(tx_clk(i)) THEN
        mux_select_arr(i) <= nxt_mux_select_arr(i);
      END IF;
    END PROCESS;

    u_dp_demux : ENTITY dp_lib.dp_demux
    GENERIC MAP (
      g_sel_ctrl_invert   => TRUE, -- We're using DOWNTO ranges in out streaming arrays, so we must invert the selection input
      g_mode              => 2,  
      g_nof_output        => c_nof_select,
      g_remove_channel_lo => FALSE, 
      g_combined          => FALSE 
    )
    PORT MAP (
      rst         => i_rx_rst(i),
      clk         => i_rx_clk(i),
  
      sel_ctrl    => TO_UINT(demux_select_arr(i)), 
  
      -- ST sinks
      snk_out     => demux_in_siso_arr(i),    
      snk_in      => demux_in_sosi_arr(i),
      -- ST source
      src_in_arr  => demux_out_siso_2arr(i), 
      src_out_arr => demux_out_sosi_2arr(i)
    );
  
    p_rx_clk : PROCESS(i_rx_rst, i_rx_clk)
    BEGIN
      IF i_rx_rst(i)='1' THEN
        demux_select_arr(i) <= (OTHERS=>'0');
      ELSIF rising_edge(i_rx_clk(i)) THEN
        demux_select_arr(i) <= nxt_demux_select_arr(i);
      END IF;
    END PROCESS;

  END GENERATE;
      
END wrap;

