--------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib, mdio_lib, technology_lib, tech_xaui_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_interface_layers_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE mdio_lib.mdio_pkg.ALL;
USE mdio_lib.mdio_vitesse_vsc8486_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;

ENTITY tr_xaui IS
  GENERIC (
    g_technology            : NATURAL := c_tech_select_default;
    g_sim                   : BOOLEAN := FALSE;
    g_sim_level             : NATURAL := 0;     -- 0 = use IP; 1 = use fast serdes model
    g_use_xgmii             : BOOLEAN := FALSE; -- Don't use streaming I/O but XGMII (e.g. conenct to 10GbE MAC)
    g_nof_xaui              : NATURAL := 1;     -- Up to 3 (hard XAUI only) supported
    g_mdio                  : BOOLEAN := FALSE;
    g_mdio_epcs_dis         : BOOLEAN := FALSE -- TRUE disables EPCS on init; e.g. to target a 10GbE card in PC that does not support it 
    --g_mdio_mm_ctrl          : BOOLEAN := FALSE  -- FALSE uses mdio_ctlr to auto execute the initialization sequence for the Vitesse chips. TRUE enables MM control.
  );
  PORT (   
    -- Transceiver PLL reference clock   
    tr_clk                  : IN  STD_LOGIC;

    -- Calibration & reconfig clock
    cal_rec_clk             : IN  STD_LOGIC;
 
    -- MM clock for register of optional MDIO master
    mm_clk                  : IN  STD_LOGIC := '0';
    mm_rst                  : IN  STD_LOGIC := '0';

    -- Streaming RX interfaces
    rx_clk                  : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
    rx_rst                  : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
    rx_sosi_arr             : OUT t_dp_sosi_arr(g_nof_xaui-1 DOWNTO 0);
    rx_siso_arr             : IN  t_dp_siso_arr(g_nof_xaui-1 DOWNTO 0) := (OTHERS=>c_dp_siso_rst);

    -- Streaming TX interfaces
    tx_clk                  : IN  STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
    tx_rst                  : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
    tx_sosi_arr             : IN  t_dp_sosi_arr(g_nof_xaui-1 DOWNTO 0) := (OTHERS=>c_dp_sosi_rst);
    tx_siso_arr             : OUT t_dp_siso_arr(g_nof_xaui-1 DOWNTO 0);

    -- Direct XGMII interface
    xgmii_tx_dc_arr         : IN  t_xgmii_dc_arr(g_nof_xaui-1 DOWNTO 0) := (OTHERS=>(OTHERS=>'0'));
    xgmii_rx_dc_arr         : OUT t_xgmii_dc_arr(g_nof_xaui-1 DOWNTO 0);
     
    --Serial I/O
    xaui_rx                 : IN  t_xaui_arr(g_nof_xaui-1 DOWNTO 0);
    xaui_tx                 : OUT t_xaui_arr(g_nof_xaui-1 DOWNTO 0);   

    -- XAUI PHY IP MM control/status
    xaui_mosi               : IN  t_mem_mosi := c_mem_mosi_rst;
    xaui_miso               : OUT t_mem_miso; 
      
    -- MDIO master = mm slave
    mdio_mosi_arr           : IN  t_mem_mosi_arr(g_nof_xaui-1 DOWNTO 0) := (OTHERS=>c_mem_mosi_rst);
    mdio_miso_arr           : OUT t_mem_miso_arr(g_nof_xaui-1 DOWNTO 0);

    -- MDIO External clock and serial data.
    mdio_rst                : OUT STD_LOGIC;
    mdio_mdc                : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);       
    mdio_mdat_in            : IN  STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0) := (OTHERS=>'0');
    mdio_mdat_oen           : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0)
  );
END tr_xaui;


ARCHITECTURE str OF tr_xaui IS  

  SIGNAL crc_rx_ready              : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0); --crc = synchronous to Cal_Rec_Clk
  SIGNAL crc_tx_ready              : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);

  SIGNAL rxc_rx_ready              : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
  SIGNAL txc_tx_ready              : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);

  SIGNAL i_rx_rst                  : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
  SIGNAL i_tx_rst                  : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);

  SIGNAL tr_rst                    : STD_LOGIC;
  SIGNAL i_rx_clk                  : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);

  --XGMII data and control combined:
  SIGNAL xgmii_tx_dc               : t_xgmii_dc_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL xgmii_rx_dc               : t_xgmii_dc_arr(g_nof_xaui-1 DOWNTO 0);

  --XGMII control bits (one for each XGMII lane):
  SIGNAL xgmii_tx_c                : t_xgmii_c_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL xgmii_rx_c                : t_xgmii_c_arr(g_nof_xaui-1 DOWNTO 0);
  
  --XGMII data
  SIGNAL xgmii_tx_d                : t_xgmii_d_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL xgmii_rx_d                : t_xgmii_d_arr(g_nof_xaui-1 DOWNTO 0);

  SIGNAL a_rx_channelaligned       : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
--  SIGNAL rxc_rx_channelaligned     : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
  SIGNAL txc_rx_channelaligned_dly : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0); 

  -- MDIO (optional) signals
  SIGNAL mdio_hdr                  : t_slv_16_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL mdio_tx_dat               : t_slv_16_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL mdio_rx_dat               : t_slv_16_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL mdio_en_evt               : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
  SIGNAL mdio_done                 : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
  SIGNAL mdio_done_ack_evt         : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);

  SIGNAL reg_mdio_hdr              : t_slv_16_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL reg_mdio_tx_dat           : t_slv_16_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL reg_mdio_rx_dat           : t_slv_16_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL reg_mdio_en_evt           : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
  SIGNAL reg_mdio_done             : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
  SIGNAL reg_mdio_done_ack_evt     : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);

  SIGNAL ctlr_mdio_rst             : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
  SIGNAL ctlr_mdio_hdr             : t_slv_16_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL ctlr_mdio_tx_dat          : t_slv_16_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL ctlr_mdio_rx_dat          : t_slv_16_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL ctlr_mdio_en_evt          : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
  SIGNAL ctlr_mdio_done            : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
  SIGNAL ctlr_mdio_done_ack_evt    : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);

  SIGNAL ctlr_exec_complete        : STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);

  SIGNAL tx_framer_siso_arr        : t_dp_siso_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL tx_framer_sosi_arr        : t_dp_sosi_arr(g_nof_xaui-1 DOWNTO 0);

BEGIN

  -----------------------------------------------------------------------------
  -- XAUI PHY
  -----------------------------------------------------------------------------
  gen_phy: IF g_sim = FALSE OR g_sim_level = 0 GENERATE
    -- Altera's IP
    u_tech_xaui : ENTITY tech_xaui_lib.tech_xaui
    GENERIC MAP (
      g_technology => g_technology,
      g_nof_xaui   => g_nof_xaui 
    )
    PORT MAP (
      tr_clk                  => tr_clk,
 
      mm_rst                  => mm_rst,  
      mm_clk                  => mm_clk,
 
      cal_rec_clk             => cal_rec_clk,
    
      tx_clk_arr              => tx_clk,
      rx_clk_arr              => i_rx_clk,
     
      crc_rx_ready_arr        => crc_rx_ready,
      crc_tx_ready_arr        => crc_tx_ready,
  
      a_rx_channelaligned_arr => a_rx_channelaligned,
  
      xgmii_tx_dc_arr         => xgmii_tx_dc,
      xgmii_rx_dc_arr         => xgmii_rx_dc,
  
      xaui_rx_arr             => xaui_rx,
      xaui_tx_arr             => xaui_tx,

      xaui_mosi               => xaui_mosi,
      xaui_miso               => xaui_miso
    );
  END GENERATE;  

  gen_sim: IF g_sim = TRUE AND g_sim_level = 1 GENERATE
    -- Behavioural serdes model (fast)
    u_sim_xaui : ENTITY work.sim_xaui
    GENERIC MAP (
      g_nof_xaui => g_nof_xaui 
    )
    PORT MAP (
      tr_clk                  => tr_clk,
      tr_rst                  => tr_rst,
  
      cal_rec_clk             => cal_rec_clk,
    
      tx_clk_arr              => tx_clk,
      rx_clk_arr              => i_rx_clk,
     
      crc_rx_ready_arr        => crc_rx_ready,
      crc_tx_ready_arr        => crc_tx_ready,
  
      a_rx_channelaligned_arr => a_rx_channelaligned,
  
      xgmii_tx_dc_arr         => xgmii_tx_dc,
      xgmii_rx_dc_arr         => xgmii_rx_dc,
  
      xaui_rx_arr             => xaui_rx,
      xaui_tx_arr             => xaui_tx
    );
  END GENERATE;  

  -----------------------------------------------------------------------------
  -- Clock and reset generation
  -----------------------------------------------------------------------------
  u_areset_tr_rst : ENTITY common_lib.common_areset
  GENERIC MAP(
    g_rst_level => '1',
    g_delay_len => 4
  )
  PORT MAP(
    clk     => tr_clk,
    in_rst  => '0',
    out_rst => tr_rst
  );    
 
  gen_nof_xaui : FOR i IN g_nof_xaui-1 DOWNTO 0 GENERATE

    tx_rst(i) <= i_tx_rst(i);

    rx_clk(i) <= i_rx_clk(i);
    rx_rst(i) <= i_rx_rst(i);

    i_rx_rst(i) <= NOT rxc_rx_ready(i);
    i_tx_rst(i) <= NOT txc_tx_ready(i); 

    u_async_rxc_rx_ready: ENTITY common_lib.common_async
    GENERIC MAP(
      g_rst_level => '0'
    )
    PORT MAP(
      clk  => i_rx_clk(i),
      din  => crc_rx_ready(i),
      dout => rxc_rx_ready(i)
    );

    u_async_txc_tx_ready: ENTITY common_lib.common_async
    GENERIC MAP(
      g_rst_level => '0'
    )
    PORT MAP(
      clk  => tx_clk(i),
      din  => crc_tx_ready(i),
      dout => txc_tx_ready(i)
    );

    -----------------------------------------------------------------------------
    -- SOSI <-> XGMII
    -----------------------------------------------------------------------------
    gen_sosi_io: IF g_use_xgmii=FALSE GENERATE
 
      xgmii_tx_dc(i) <= func_xgmii_dc(xgmii_tx_d(i), xgmii_tx_c(i));
  
      xgmii_rx_d(i) <= func_xgmii_d(xgmii_rx_dc(i));
      xgmii_rx_c(i) <= func_xgmii_c(xgmii_rx_dc(i));
  
      tx_siso_arr(i).ready <= '0' WHEN txc_rx_channelaligned_dly(i) = '0' ELSE tx_framer_siso_arr(i).ready;
      tx_siso_arr(i).xon   <= txc_rx_channelaligned_dly(i);
  
      tx_framer_sosi_arr(i) <= tx_sosi_arr(i);
  
      u_tx_framer : ENTITY work.tr_xaui_framer
      GENERIC MAP (
        g_dat_len => sel_a_b(g_sim, 100, 100000),
        g_gap_len => 10
       )
      PORT MAP (
        tx_rst     => i_tx_rst(i),
        tx_clk     => tx_clk(i),
  
        snk_out    => tx_framer_siso_arr(i),
        snk_in     => tx_framer_sosi_arr(i),
  
        xgmii_tx_d => xgmii_tx_d(i),
        xgmii_tx_c => xgmii_tx_c(i)    
      );
  
      u_rx_deframer : ENTITY work.tr_xaui_deframer
      PORT MAP (
        rx_rst     => i_rx_rst(i),
        rx_clk     => i_rx_clk(i),
  
        xgmii_rx_d => xgmii_rx_d(i),
        xgmii_rx_c => xgmii_rx_c(i),
  
        src_out    => rx_sosi_arr(i)
      );

      u_tr_xaui_align_dly: ENTITY work.tr_xaui_align_dly
      GENERIC MAP(
        g_sim => g_sim
      )
      PORT MAP(
        tx_rst                    => i_tx_rst(i),
        tx_clk                    => tx_clk(i),
  
        a_rx_channelaligned       => a_rx_channelaligned(i),
        txc_rx_channelaligned_dly => txc_rx_channelaligned_dly(i)
      );

    END GENERATE;

    -----------------------------------------------------------------------------
    -- Direct XGMII
    -----------------------------------------------------------------------------
    gen_xgmii_io: IF g_use_xgmii=TRUE GENERATE
      xgmii_tx_dc(i)     <= xgmii_tx_dc_arr(i);
      xgmii_rx_dc_arr(i) <= xgmii_rx_dc(i);
    END GENERATE;

    gen_mdio: IF g_mdio = TRUE GENERATE

      -- UniBoard FN resets 4 Vitesse chips using one output.
      mdio_rst <= ctlr_mdio_rst(0);

      -----------------------------------------------------------------------------
      -- MDIO: the PHY core itself
      -----------------------------------------------------------------------------
      u_mdio_phy : ENTITY mdio_lib.mdio_phy
      GENERIC MAP(
        g_add_mdc_cycle => TRUE
      )
      PORT MAP (
        gs_sim            => g_sim,
        
        rst               => tr_rst,
        clk               => tr_clk,
      
        mdio_en_evt       => mdio_en_evt(i),
        mdio_done         => mdio_done(i),
        mdio_done_ack_evt => mdio_done_ack_evt(i),
      
        hdr               => mdio_hdr(i),
        tx_dat            => mdio_tx_dat(i),
        rx_dat            => mdio_rx_dat(i),
      
        -- External clock and serial data
        mdc               => mdio_mdc(i),
        mdat_in           => mdio_mdat_in(i),
        mdat_oen          => mdio_mdat_oen(i)
      );

      -----------------------------------------------------------------------------
      -- MDIO: MM port
      -----------------------------------------------------------------------------
      u_mdio_phy_reg : ENTITY mdio_lib.mdio_phy_reg
      PORT MAP ( 
        mm_rst            => mm_rst,
        mm_clk            => mm_clk,
    
        mdio_rst          => tr_rst,
        mdio_clk          => tr_clk,
    
        sla_in            => mdio_mosi_arr(i),
        sla_out           => mdio_miso_arr(i),
    
        mdio_en_evt       => reg_mdio_en_evt(i),
        mdio_done_ack_evt => reg_mdio_done_ack_evt(i),
        mdio_done         => reg_mdio_done(i),
    
        mdio_hdr          => reg_mdio_hdr(i),
        mdio_tx_dat       => reg_mdio_tx_dat(i),
        mdio_rx_dat       => reg_mdio_rx_dat(i)
      );

      -----------------------------------------------------------------------------
      -- MDIO controller auto-executes MDIO sequence on startup
      -----------------------------------------------------------------------------
      gen_mdio_epcs : IF g_mdio_epcs_dis = FALSE GENERATE
        u_mdio_ctlr_epcs : ENTITY mdio_lib.mdio_ctlr
        GENERIC MAP (
           g_mdio_prtad           => c_mdio_vsc8486_prtad,
           g_mdio_cmd_arr         => c_mdio_vsc8486_init_cmd_arr,
           g_mdio_rst_level       => '0', 
           g_mdio_rst_cycles      => 250000,           
           g_mdio_post_rst_cycles => 250000
          ) 
        PORT MAP (
          rst               => tr_rst,
          clk               => tr_clk,

          mdio_rst          => ctlr_mdio_rst(i),
          mdio_en_evt       => ctlr_mdio_en_evt(i),
          mdio_done         => ctlr_mdio_done(i),
      
          mdio_done_ack_evt => ctlr_mdio_done_ack_evt(i),
      
          hdr               => ctlr_mdio_hdr(i),
          tx_dat            => ctlr_mdio_tx_dat(i),

          exec_complete     => ctlr_exec_complete(i)
        );
      END GENERATE;

      -----------------------------------------------------------------------------
      -- MDIO controller auto-executes MDIO sequence on startup
      -----------------------------------------------------------------------------
      gen_mdio_no_epcs : IF g_mdio_epcs_dis = TRUE GENERATE
        u_mdio_ctlr_no_epcs : ENTITY mdio_lib.mdio_ctlr
        GENERIC MAP (
           g_mdio_prtad           => c_mdio_vsc8486_prtad,
           g_mdio_cmd_arr         => c_mdio_vsc8486_init_epcs_dis_cmd_arr,
           g_mdio_rst_level       => '0', 
           g_mdio_rst_cycles      => 250000,
           g_mdio_post_rst_cycles => 250000
          ) 
        PORT MAP (
          rst               => tr_rst,
          clk               => tr_clk,

          mdio_rst          => ctlr_mdio_rst(i),      
          mdio_en_evt       => ctlr_mdio_en_evt(i),
          mdio_done         => ctlr_mdio_done(i),
      
          mdio_done_ack_evt => ctlr_mdio_done_ack_evt(i),
      
          hdr               => ctlr_mdio_hdr(i),
          tx_dat            => ctlr_mdio_tx_dat(i),

          exec_complete     => ctlr_exec_complete(i)
        );
      END GENERATE;

      -----------------------------------------------------------------------------
      -- MDIO: Connect the mdio_ctlr to the mdio_phy initially, when it's done
      --       connect the MM controller to allow user control/monitoring.
      -----------------------------------------------------------------------------
      mdio_en_evt(i)       <= ctlr_mdio_en_evt(i)       WHEN ctlr_exec_complete(i)='0' ELSE reg_mdio_en_evt(i); 
      mdio_done_ack_evt(i) <= ctlr_mdio_done_ack_evt(i) WHEN ctlr_exec_complete(i)='0' ELSE reg_mdio_done_ack_evt(i);
      mdio_hdr(i)          <= ctlr_mdio_hdr(i)          WHEN ctlr_exec_complete(i)='0' ELSE reg_mdio_hdr(i);
      mdio_tx_dat(i)       <= ctlr_mdio_tx_dat(i)       WHEN ctlr_exec_complete(i)='0' ELSE reg_mdio_tx_dat(i);       
    
      ctlr_mdio_done(i)    <= mdio_done(i)   WHEN ctlr_exec_complete(i)='0' ELSE '0';
      reg_mdio_done(i)     <= mdio_done(i)   WHEN ctlr_exec_complete(i)='1' ELSE '0';
      reg_mdio_rx_dat(i)   <= mdio_rx_dat(i) WHEN ctlr_exec_complete(i)='1' ELSE (OTHERS=>'0');


    END GENERATE; -- if g_mdio = True

  END GENERATE; -- for i in 0..g_nof_xaui
      
END str;

