-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE common_lib.common_interface_layers_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE dp_lib.tb_dp_pkg.ALL;

ENTITY tb_tr_xaui_deframer IS
END tb_tr_xaui_deframer;

ARCHITECTURE tb OF tb_tr_xaui_deframer IS

  CONSTANT c_clk_period   : TIME    := 10 ns;   
  CONSTANT c_rl           : NATURAL := 1;
  CONSTANT c_tx_init      : NATURAL := 0;

  SIGNAL rst              : STD_LOGIC;
  SIGNAL clk              : STD_LOGIC := '1';
  
  SIGNAL tx_enable        : STD_LOGIC;

  SIGNAL tx_siso          : t_dp_siso;
  SIGNAL tx_sosi          : t_dp_sosi;

  SIGNAL gap_siso         : t_dp_siso;
  SIGNAL gap_sosi         : t_dp_sosi;

  SIGNAL xgmii_tx_d       : STD_LOGIC_VECTOR(c_xgmii_data_w-1 DOWNTO 0);
  SIGNAL xgmii_tx_c       : STD_LOGIC_VECTOR(c_xgmii_nof_lanes-1 DOWNTO 0);

  SIGNAL xgmii_rx_d       : STD_LOGIC_VECTOR(c_xgmii_data_w-1 DOWNTO 0);
  SIGNAL xgmii_rx_c       : STD_LOGIC_VECTOR(c_xgmii_nof_lanes-1 DOWNTO 0);

  SIGNAL deframer_sosi    : t_dp_sosi;
  SIGNAL deframer_siso    : t_dp_siso;

  SIGNAL verify_en        : STD_LOGIC;
  SIGNAL prev_data1       : STD_LOGIC_VECTOR(c_xgmii_data_w-1 DOWNTO 0);
  SIGNAL prev_data2       : STD_LOGIC_VECTOR(c_xgmii_data_w-1 DOWNTO 0);

  SIGNAL xgmii_rx_d_hi      : STD_LOGIC_VECTOR(c_xgmii_data_w/2-1 DOWNTO 0);
  SIGNAL xgmii_rx_d_lo      : STD_LOGIC_VECTOR(c_xgmii_data_w/2-1 DOWNTO 0);

  SIGNAL xgmii_tx_d_hi      : STD_LOGIC_VECTOR(c_xgmii_data_w/2-1 DOWNTO 0);
  SIGNAL xgmii_tx_d_lo      : STD_LOGIC_VECTOR(c_xgmii_data_w/2-1 DOWNTO 0);

  SIGNAL prev_xgmii_tx_d_hi : STD_LOGIC_VECTOR(c_xgmii_data_w/2-1 DOWNTO 0);
  SIGNAL prev_xgmii_tx_c_hi : STD_LOGIC_VECTOR(c_xgmii_nof_lanes/2-1 DOWNTO 0);

  SIGNAL xgmii_tx_c_hi      : STD_LOGIC_VECTOR(c_xgmii_nof_lanes/2-1 DOWNTO 0);
  SIGNAL xgmii_tx_c_lo      : STD_LOGIC_VECTOR(c_xgmii_nof_lanes/2-1 DOWNTO 0);

  SIGNAL xgmii_rx_c_hi      : STD_LOGIC_VECTOR(c_xgmii_nof_lanes/2-1 DOWNTO 0);
  SIGNAL xgmii_rx_c_lo      : STD_LOGIC_VECTOR(c_xgmii_nof_lanes/2-1 DOWNTO 0);

  SIGNAL misalign           : STD_LOGIC;

BEGIN

  rst <= '1', '0' AFTER c_clk_period*7;
  clk <= NOT clk AFTER c_clk_period/2;

  xgmii_tx_d_hi <= xgmii_tx_d(c_xgmii_data_w  -1 DOWNTO c_xgmii_data_w/2);
  xgmii_tx_d_lo <= xgmii_tx_d(c_xgmii_data_w/2-1 DOWNTO 0);

  xgmii_tx_c_hi <= xgmii_tx_c(c_xgmii_nof_lanes  -1 DOWNTO c_xgmii_nof_lanes/2);
  xgmii_tx_c_lo <= xgmii_tx_c(c_xgmii_nof_lanes/2-1 DOWNTO 0);

  xgmii_rx_d <= xgmii_rx_d_hi & xgmii_rx_d_lo;
  xgmii_rx_c <= xgmii_rx_c_hi & xgmii_rx_c_lo;

  p_stim : PROCESS
  BEGIN
    tx_enable <= '0';
    verify_en <= '0';   
    misalign  <= '0';

    WAIT FOR c_clk_period*100; 
    tx_enable <= '1';
    gap_siso.ready <= '1';
  
    WAIT FOR c_clk_period*20; 
    verify_en <= '1';

    -- Let run for a while with aligned data
    WAIT FOR 3 us;

    WAIT UNTIL deframer_sosi.valid = '0';
    WAIT UNTIL rising_edge(clk);
    misalign  <= '1'; -- Now change the 32b word boundary

    WAIT;
  END PROCESS;  
  
   -- Generate tx_sosi for DUT using counter data generator
  proc_dp_gen_data(c_rl, c_xgmii_data_w, c_tx_init, rst, clk, tx_enable, tx_siso, tx_sosi);
  
  -- Introduce some gaps in the streaming data to make sure framer handles them correctly.
  u_dp_gap : ENTITY dp_lib.dp_gap
  GENERIC MAP (
    g_dat_len    => 40,
    g_gap_len    => 20
  )
  PORT MAP (
    rst       => rst,
    clk       => clk,

    snk_out   => tx_siso,
    snk_in    => tx_sosi,
    
    src_in    => gap_siso,
    src_out   => gap_sosi
  );

  u_framer : ENTITY work.tr_xaui_framer
  GENERIC MAP (
    g_dat_len => 100,
    g_gap_len => 5
   )
  PORT MAP (
    tx_rst     => rst,
    tx_clk     => clk,

    snk_out    => gap_siso,
    snk_in     => gap_sosi,

    xgmii_tx_d => xgmii_tx_d,
    xgmii_tx_c => xgmii_tx_c    
  );

  -- ============== We'll emulate misalignment here ===========================
 
  -- Aligned data:
  -- RX Hi:  0 1 2 3 4 5
  -- RX Lo:  0 1 2 3 4 5
  xgmii_rx_d_hi <= xgmii_tx_d_lo WHEN misalign = '1' ELSE xgmii_tx_d_hi;
  xgmii_rx_c_hi <= xgmii_tx_c_lo WHEN misalign = '1' ELSE xgmii_tx_c_hi;

  -- Misligned data:
  -- RX Hi:  1 2 3 4 5 6
  -- RX Lo:  0 1 2 3 4 5
  xgmii_rx_d_lo <= prev_xgmii_tx_d_hi WHEN misalign = '1' ELSE xgmii_tx_d_lo;
  xgmii_rx_c_lo <= prev_xgmii_tx_c_hi WHEN misalign = '1' ELSE xgmii_tx_c_lo;

  p_clk : PROCESS(clk, rst)
  BEGIN
    IF rst='1' THEN
      prev_xgmii_tx_d_hi <= x"07070707";
      prev_xgmii_tx_c_hi <= x"F";
    ELSIF rising_edge(clk) THEN
      prev_xgmii_tx_d_hi <= xgmii_tx_d_hi;
      prev_xgmii_tx_c_hi <= xgmii_tx_c_hi;
    END IF;
  END PROCESS;

  -- ===========================================================================

  dut : ENTITY work.tr_xaui_deframer
  PORT MAP (
    rx_rst     => rst,
    rx_clk     => clk,

    xgmii_rx_d => xgmii_rx_d,
    xgmii_rx_c => xgmii_rx_c,

    src_out    => deframer_sosi

  );

  -- Verify DUT output incrementing data, prev_data is an auxiliary signal needed by the proc
  proc_common_verify_data(c_rl, clk, verify_en, deframer_siso.ready, deframer_sosi.valid, deframer_sosi.data(c_xgmii_data_w-1 DOWNTO 0), prev_data1);
  proc_dp_verify_data("data", c_rl, clk, verify_en, deframer_siso.ready, deframer_sosi.valid, deframer_sosi.data(c_xgmii_data_w-1 DOWNTO 0), prev_data2);  
 
END tb;
