--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- Declare IP libraries to ensure default binding in simulation. The IP library clause is ignored by synthesis.
LIBRARY ip_arria10_phy_10gbase_r_lib;
LIBRARY ip_arria10_transceiver_pll_156_312_lib;
LIBRARY ip_arria10_transceiver_pll_10g_lib;
LIBRARY ip_arria10_transceiver_reset_controller_1_lib;

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_interface_layers_pkg.ALL;
USE work.tech_10gbase_r_component_pkg.ALL;

ENTITY tech_10gbase_r_arria10 IS
  GENERIC (
    g_sim            : BOOLEAN := FALSE;
    g_nof_channels   : NATURAL := 1
  );
  PORT (
    -- Transceiver PLL reference clock
    tr_ref_clk              : IN  STD_LOGIC;    -- 644.531250 MHz
    
    -- Derived clocks
    clk_312_out             : OUT STD_LOGIC;    -- 312.5 MHz
    clk_156_out             : OUT STD_LOGIC;    -- 156.25 MHz
    clk_156_in              : IN  STD_LOGIC;    -- externally connect to clk_156_out to avoid delta-cycle differences in simulation
    rst_156                 : OUT STD_LOGIC;    -- reset for clk_156_in clock domain

    -- XGMII interface
    xgmii_tx_dc_arr         : IN  t_xgmii_dc_arr(g_nof_channels-1 DOWNTO 0);  -- 72 bit
    xgmii_rx_dc_arr         : OUT t_xgmii_dc_arr(g_nof_channels-1 DOWNTO 0);  -- 72 bit

    -- PHY serial IO
    tx_serial_arr           : OUT STD_LOGIC_VECTOR(g_nof_channels-1 DOWNTO 0);
    rx_serial_arr           : IN  STD_LOGIC_VECTOR(g_nof_channels-1 DOWNTO 0)
  );
END tech_10gbase_r_arria10;


ARCHITECTURE str OF tech_10gbase_r_arria10 IS

  SIGNAL tx_serial_clk          : STD_LOGIC_VECTOR(0 DOWNTO 0);

  SIGNAL tx_coreclkin           : STD_LOGIC_VECTOR(0 DOWNTO 0);
  SIGNAL rx_coreclkin           : STD_LOGIC_VECTOR(0 DOWNTO 0);

  SIGNAL tx_parallel_data_arr   : t_xgmii_d_arr(g_nof_channels-1 DOWNTO 0);  -- 64 bit
  SIGNAL rx_parallel_data_arr   : t_xgmii_d_arr(g_nof_channels-1 DOWNTO 0);  -- 64 bit
  SIGNAL tx_control_arr         : t_xgmii_c_arr(g_nof_channels-1 DOWNTO 0);  --  8 bit
  SIGNAL rx_control_arr         : t_xgmii_c_arr(g_nof_channels-1 DOWNTO 0);  --  8 bit

  -- transceiver reset controller
  SIGNAL tx_analogreset_arr     : STD_LOGIC_VECTOR(g_nof_channels-1 DOWNTO 0);
  SIGNAL tx_digitalreset_arr    : STD_LOGIC_VECTOR(g_nof_channels-1 DOWNTO 0);
  SIGNAL rx_analogreset_arr     : STD_LOGIC_VECTOR(g_nof_channels-1 DOWNTO 0);
  SIGNAL rx_digitalreset_arr    : STD_LOGIC_VECTOR(g_nof_channels-1 DOWNTO 0);
  SIGNAL tx_cal_busy_arr        : STD_LOGIC_VECTOR(g_nof_channels-1 DOWNTO 0);
  SIGNAL rx_cal_busy_arr        : STD_LOGIC_VECTOR(g_nof_channels-1 DOWNTO 0);
  SIGNAL rx_is_lockedtodata_arr : STD_LOGIC_VECTOR(g_nof_channels-1 DOWNTO 0);

  SIGNAL cal_busy_arr           : STD_LOGIC_VECTOR(g_nof_channels-1 DOWNTO 0);

  -- transceiver ATX PLL for 10G
  SIGNAL atx_pll_powerdown      : STD_LOGIC_VECTOR(g_nof_channels-1 DOWNTO 0);  -- only use reset controller 0 for ATX PLL power down
  SIGNAL atx_pll_locked_arr     : STD_LOGIC_VECTOR(g_nof_channels-1 DOWNTO 0);
  SIGNAL atx_pll_locked         : STD_LOGIC;
  SIGNAL atx_pll_cal_busy       : STD_LOGIC;
  
  -- transceiver fractional PLL for 156.25 MHz and 312.5 MHz from tr_ref_clk = 644.53125 MHz
  SIGNAL fpll_cal_busy          : STD_LOGIC;
  SIGNAL fpll_locked            : STD_LOGIC;
  SIGNAL fpll_locked_n          : STD_LOGIC;
  SIGNAL i_rst_156              : STD_LOGIC;
  
BEGIN

  -- Clocks
  tx_coreclkin(0) <= clk_156_in;
  rx_coreclkin(0) <= clk_156_in;
                                
  gen_phy : FOR I IN 0 TO g_nof_channels-1 GENERATE                                
    -- Reset controller
    cal_busy_arr(I) <= tx_cal_busy_arr(I) OR atx_pll_cal_busy;
    
    -- On hardware use atx_pll_locked for all channels, in simulation model some timing difference between the channels
    on_hw : IF g_sim=FALSE GENERATE
      atx_pll_locked_arr(I) <= atx_pll_locked;
    END GENERATE;
    in_sim : IF g_sim=TRUE GENERATE
      atx_pll_locked_arr(I) <= TRANSPORT atx_pll_locked AFTER tech_10gbase_r_clk_156_period*I;
    END GENERATE;
    
    -- XGMII
    tx_parallel_data_arr(I) <= func_xgmii_d(xgmii_tx_dc_arr(I));
    tx_control_arr(I)       <= func_xgmii_c(xgmii_tx_dc_arr(I));
  
    xgmii_rx_dc_arr(I)      <= func_xgmii_dc(rx_parallel_data_arr(I), rx_control_arr(I));
  
    u_ip_arria10_phy_10gbase_r_top : ip_arria10_phy_10gbase_r_top
    PORT MAP (
      tx_analogreset          => tx_analogreset_arr(I DOWNTO I),
      tx_digitalreset         => tx_digitalreset_arr(I DOWNTO I),
      rx_analogreset          => rx_analogreset_arr(I DOWNTO I),
      rx_digitalreset         => rx_digitalreset_arr(I DOWNTO I),
      tx_cal_busy             => tx_cal_busy_arr(I DOWNTO I),
      rx_cal_busy             => rx_cal_busy_arr(I DOWNTO I),
  
      tx_serial_clk0          => tx_serial_clk,
      rx_cdr_refclk0          => tr_ref_clk,
      tx_serial_data          => tx_serial_arr(I DOWNTO I),
      rx_serial_data          => rx_serial_arr(I DOWNTO I),
  
      rx_is_lockedtoref       => OPEN,
      rx_is_lockedtodata      => rx_is_lockedtodata_arr(I DOWNTO I),
  
      tx_coreclkin            => tx_coreclkin,     -- 156.25 MHz
      rx_coreclkin            => rx_coreclkin,     -- 156.25 MHz
  
      tx_parallel_data        => tx_parallel_data_arr(I),
      rx_parallel_data        => rx_parallel_data_arr(I),
      tx_control              => tx_control_arr(I),
  
      --tx_err_ins              : in  std_logic                     := '0';             --              tx_err_ins.tx_err_ins
      --unused_tx_parallel_data : in  std_logic_vector(63 downto 0) := (others => '0'); -- unused_tx_parallel_data.unused_tx_parallel_data
      --unused_tx_control       : in  std_logic_vector(8 downto 0)  := (others => '0'); --       unused_tx_control.unused_tx_control
  
      rx_control              => rx_control_arr(I)
  
      --unused_rx_parallel_data : out std_logic_vector(63 downto 0);                    -- unused_rx_parallel_data.unused_rx_parallel_data
      --unused_rx_control       : out std_logic_vector(11 downto 0);                    --       unused_rx_control.unused_rx_control
      --tx_enh_data_valid       : in  std_logic_vector(0 downto 0)  := (others => '0'); --       tx_enh_data_valid.tx_enh_data_valid
      --tx_enh_fifo_full        : out std_logic_vector(0 downto 0);                     --        tx_enh_fifo_full.tx_enh_fifo_full
      --tx_enh_fifo_pfull       : out std_logic_vector(0 downto 0);                     --       tx_enh_fifo_pfull.tx_enh_fifo_pfull
      --tx_enh_fifo_empty       : out std_logic_vector(0 downto 0);                     --       tx_enh_fifo_empty.tx_enh_fifo_empty
      --tx_enh_fifo_pempty      : out std_logic_vector(0 downto 0);                     --      tx_enh_fifo_pempty.tx_enh_fifo_pempty
      --rx_enh_data_valid       : out std_logic_vector(0 downto 0);                     --       rx_enh_data_valid.rx_enh_data_valid
      --rx_enh_fifo_full        : out std_logic_vector(0 downto 0);                     --        rx_enh_fifo_full.rx_enh_fifo_full
      --rx_enh_fifo_empty       : out std_logic_vector(0 downto 0);                     --       rx_enh_fifo_empty.rx_enh_fifo_empty
      --rx_enh_fifo_del         : out std_logic_vector(0 downto 0);                     --         rx_enh_fifo_del.rx_enh_fifo_del
      --rx_enh_fifo_insert      : out std_logic_vector(0 downto 0);                     --      rx_enh_fifo_insert.rx_enh_fifo_insert
      --rx_enh_highber          : out std_logic_vector(0 downto 0);                     --          rx_enh_highber.rx_enh_highber
      --rx_enh_blk_lock         : out std_logic_vector(0 downto 0);                     --         rx_enh_blk_lock.rx_enh_blk_lock
    );
  
    u_ip_arria10_transceiver_reset_controller_1_top : ip_arria10_transceiver_reset_controller_1_top
    PORT MAP (
      clock              => clk_156_in,
      reset              => i_rst_156,
      pll_powerdown      => atx_pll_powerdown(I DOWNTO I),
      tx_analogreset     => tx_analogreset_arr(I DOWNTO I),
      tx_digitalreset    => tx_digitalreset_arr(I DOWNTO I),
      tx_ready           => OPEN,
      pll_locked         => atx_pll_locked_arr(I DOWNTO I),
      pll_select         => "0",                   -- set to zero when using one PLL
      tx_cal_busy        => cal_busy_arr(I DOWNTO I),
      rx_analogreset     => rx_analogreset_arr(I DOWNTO I),
      rx_digitalreset    => rx_digitalreset_arr(I DOWNTO I),
      rx_ready           => OPEN,
      rx_is_lockedtodata => rx_is_lockedtodata_arr(I DOWNTO I),
      rx_cal_busy        => rx_cal_busy_arr(I DOWNTO I)
    );
  END GENERATE;

  -- ATX PLL
  u_ip_arria10_transceiver_pll_10g_top : ip_arria10_transceiver_pll_10g_top
  PORT MAP (
    pll_powerdown => atx_pll_powerdown(0),
    pll_refclk0   => tr_ref_clk,
    tx_serial_clk => tx_serial_clk(0),
    pll_locked    => atx_pll_locked,
    pll_cal_busy  => atx_pll_cal_busy
  );
  
  -- fractional PLL
  u_ip_arria10_transceiver_pll_156_312_top : ip_arria10_transceiver_pll_156_312_top
  PORT MAP (
		pll_refclk0   => tr_ref_clk,
		pll_powerdown => '0',
		pll_locked    => fpll_locked,
		outclk0       => clk_156_out,
		pll_cal_busy  => fpll_cal_busy,
		outclk1       => clk_312_out
  );
  
  fpll_locked_n <= NOT fpll_locked;
  
  u_common_areset : ENTITY common_lib.common_areset
  GENERIC MAP (
    g_rst_level => '1',
    g_delay_len => c_meta_delay_len
  )
  PORT MAP (
    in_rst    => fpll_locked_n,
    clk       => clk_156_in,
    out_rst   => i_rst_156
  );
  
  rst_156 <= i_rst_156;
  
END str;
