-------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, technology_lib, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.tb_common_mem_pkg.ALL;
USE common_lib.common_network_layers_pkg.ALL;
USE common_lib.common_network_total_header_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE dp_lib.tb_dp_pkg.ALL;
USE technology_lib.technology_pkg.ALL;
USE work.tech_mac_10g_component_pkg.ALL;


PACKAGE tb_tech_mac_10g_pkg IS

  -- Test bench supported packet data types
  CONSTANT c_tb_tech_mac_10g_data_type_symbols : NATURAL := 0;
  CONSTANT c_tb_tech_mac_10g_data_type_counter : NATURAL := 1;
  CONSTANT c_tb_tech_mac_10g_data_type_arp     : NATURAL := 2;
  CONSTANT c_tb_tech_mac_10g_data_type_ping    : NATURAL := 3;  -- over IP/ICMP
  CONSTANT c_tb_tech_mac_10g_data_type_udp     : NATURAL := 4;  -- over IP
  
  FUNCTION func_tech_mac_10g_header_size(data_type : NATURAL) RETURN NATURAL;  -- raw ethernet: 4 header words, protocol ethernet: 11 header words
  
  -- Configure the 10G MAC
  PROCEDURE proc_tech_mac_10g_setup(CONSTANT c_technology        : IN  NATURAL;
                                    CONSTANT src_mac             : IN  STD_LOGIC_VECTOR(c_network_eth_mac_slv'RANGE);
                                    SIGNAL   mm_clk              : IN  STD_LOGIC;
                                    SIGNAL   mm_miso             : IN  t_mem_miso;
                                    SIGNAL   mm_mosi             : OUT t_mem_mosi);
  
  PROCEDURE proc_tech_mac_10g_setup_stratixiv(CONSTANT src_mac             : IN  STD_LOGIC_VECTOR(c_network_eth_mac_slv'RANGE);
                                              SIGNAL   mm_clk              : IN  STD_LOGIC;
                                              SIGNAL   mm_miso             : IN  t_mem_miso;
                                              SIGNAL   mm_mosi             : OUT t_mem_mosi);
                                                                              
  PROCEDURE proc_tech_mac_10g_setup_arria10(CONSTANT src_mac             : IN  STD_LOGIC_VECTOR(c_network_eth_mac_slv'RANGE);
                                            SIGNAL   mm_clk              : IN  STD_LOGIC;
                                            SIGNAL   mm_miso             : IN  t_mem_miso;
                                            SIGNAL   mm_mosi             : OUT t_mem_mosi);
                                              
  PROCEDURE proc_tech_mac_10g_tx_packet(CONSTANT total_header    : IN  t_network_total_header;
                                        CONSTANT data_len        : IN  NATURAL;  -- in symbols = octets = bytes
                                        CONSTANT c_data_type     : IN  NATURAL;  -- c_tb_tech_mac_10g_data_type_*
                                        CONSTANT c_ready_latency : IN  NATURAL;  -- 0, 1 are supported by proc_dp_stream_ready_latency()
                                        CONSTANT c_nof_not_valid : IN  NATURAL;  -- when > 0 then pull tx valid low for c_nof_not_valid beats during tx
                                        SIGNAL   ff_clk          : IN  STD_LOGIC;
                                        SIGNAL   ff_en           : IN  STD_LOGIC;  -- similar purpose as c_nof_not_valid, but not used so pass on signal '1'
                                        SIGNAL   ff_src_in       : IN  t_dp_siso;
                                        SIGNAL   ff_src_out      : OUT t_dp_sosi);

  -- Receive and verify packet from the 10G MAC
  PROCEDURE proc_tech_mac_10g_rx_packet(CONSTANT total_header : IN  t_network_total_header;
                                        CONSTANT c_data_type  : IN  NATURAL;  -- c_tb_tech_mac_10g_data_type_*
                                        SIGNAL   ff_clk       : IN  STD_LOGIC;
                                        SIGNAL   ff_snk_in    : IN  t_dp_sosi;
                                        SIGNAL   ff_snk_out   : OUT t_dp_siso);

END tb_tech_mac_10g_pkg;


PACKAGE BODY tb_tech_mac_10g_pkg IS

  ------------------------------------------------------------------------------
  -- LOCAL ITEMS
  ------------------------------------------------------------------------------
  
  CONSTANT c_nof_eth_beats  : NATURAL := c_network_total_header_64b_eth_nof_words;  -- = 3 (2.5), nof words in eth part of the header
  CONSTANT c_nof_hdr_beats  : NATURAL := c_network_total_header_64b_nof_words;      -- = 6, nof words in the total header

  
  ------------------------------------------------------------------------------
  -- GLOBAL ITEMS
  ------------------------------------------------------------------------------
  
  FUNCTION func_tech_mac_10g_header_size(data_type : NATURAL) RETURN NATURAL IS
  BEGIN
    CASE data_type IS
      WHEN c_tb_tech_mac_10g_data_type_symbols => RETURN c_network_total_header_64b_eth_nof_words;
      WHEN c_tb_tech_mac_10g_data_type_counter => RETURN c_network_total_header_64b_eth_nof_words;
      WHEN OTHERS => NULL;
    END CASE;
    RETURN c_network_total_header_64b_nof_words;
  END func_tech_mac_10g_header_size;
  
  
  -- Configure the 10G MAC
  PROCEDURE proc_tech_mac_10g_setup(CONSTANT c_technology        : IN  NATURAL;
                                    CONSTANT src_mac             : IN  STD_LOGIC_VECTOR(c_network_eth_mac_slv'RANGE);
                                    SIGNAL   mm_clk              : IN  STD_LOGIC;
                                    SIGNAL   mm_miso             : IN  t_mem_miso;
                                    SIGNAL   mm_mosi             : OUT t_mem_mosi) IS
  BEGIN
    CASE c_technology IS
      WHEN c_tech_stratixiv => proc_tech_mac_10g_setup_stratixiv(src_mac, mm_clk, mm_miso, mm_mosi);
      WHEN c_tech_arria10   => proc_tech_mac_10g_setup_arria10(  src_mac, mm_clk, mm_miso, mm_mosi);
      WHEN OTHERS           => proc_tech_mac_10g_setup_stratixiv(src_mac, mm_clk, mm_miso, mm_mosi);  -- default to c_tech_stratixiv  
    END CASE;
  END proc_tech_mac_10g_setup;
  
  -- . The src_mac[47:0] = 0x123456789ABC for MAC address 12-34-56-78-9A-BC
  PROCEDURE proc_tech_mac_10g_setup_stratixiv(CONSTANT src_mac             : IN  STD_LOGIC_VECTOR(c_network_eth_mac_slv'RANGE);
                                              SIGNAL   mm_clk              : IN  STD_LOGIC;
                                              SIGNAL   mm_miso             : IN  t_mem_miso;
                                              SIGNAL   mm_mosi             : OUT t_mem_mosi) IS
    --CONSTANT c_mac0       : INTEGER := TO_SINT(hton(src_mac(31 DOWNTO  0), 4));
    --CONSTANT c_mac1       : INTEGER := TO_SINT(hton(src_mac(47 DOWNTO 32), 2));
    CONSTANT c_mac0       : INTEGER := TO_SINT(src_mac(31 DOWNTO  0));
    CONSTANT c_mac1       : INTEGER := TO_SINT(src_mac(47 DOWNTO 32));
  BEGIN
    -- MAC control
    -- . rx read
    proc_mem_mm_bus_rd(16#0040#,               mm_clk, mm_miso, mm_mosi);  -- RW, rx_padcrc_control = 0x1, remove CRC (use 0x3 to also remove padding)
    proc_mem_mm_bus_rd(16#0080#,               mm_clk, mm_miso, mm_mosi);  -- RW, rx_padcheck_control = 0x2, check CRC
    proc_mem_mm_bus_rd(16#0800#,               mm_clk, mm_miso, mm_mosi);  -- RW, rx_frame_control = 0x3, for promiscuous (transparent) mode
                                                                           --     [0] = 0 only receive SRC_MAC, 1 accept all unicast
                                                                           --     [1] = 0 drop multi cast, 1 to accept all multicast
    proc_mem_mm_bus_rd(16#0801#,               mm_clk, mm_miso, mm_mosi);  -- RW, rx_frame_maxlength = 1518 = 0x5EE
    proc_mem_mm_bus_rd(16#0802#,               mm_clk, mm_miso, mm_mosi);  -- RW, rx_frame_addr0 = 0, e.g. 0x56789ABC,
    proc_mem_mm_bus_rd(16#0803#,               mm_clk, mm_miso, mm_mosi);  -- RW, rx_frame_addr1 = 0, e.g. 0x1234, for primary SRC_MAC = 12-34-56-78-9A-BC
    proc_common_wait_some_cycles(mm_clk, 10);
    
    -- . tx read
    proc_mem_mm_bus_rd(16#1040#,               mm_clk, mm_miso, mm_mosi);  -- RW, tx_padins_control = 1, insert padding to ensure 64 bytes including CRC
    proc_mem_mm_bus_rd(16#1080#,               mm_clk, mm_miso, mm_mosi);  -- RW, tx_crcins_control = 3, compute and insert CRC
    proc_mem_mm_bus_rd(16#1120#,               mm_clk, mm_miso, mm_mosi);  -- RW, tx_unidir_control = 0, use 0x1 to enable unidirectional mode
    proc_mem_mm_bus_rd(16#1200#,               mm_clk, mm_miso, mm_mosi);  -- RW, tx_addrins_control = 0, use 0x1 to overwrite frame SRC_MAC with tx_addrins_macaddr0,1
    proc_mem_mm_bus_rd(16#1201#,               mm_clk, mm_miso, mm_mosi);  -- RW, tx_addrins_macaddr0 = 0, e.g. 0x56789ABC,
    proc_mem_mm_bus_rd(16#1202#,               mm_clk, mm_miso, mm_mosi);  -- RW, tx_addrins_macaddr1 = 0, e.g. 0x1234, for frame SRC_MAC = 12-34-56-78-9A-BC
    proc_mem_mm_bus_rd(16#1801#,               mm_clk, mm_miso, mm_mosi);  -- RW, tx_frame_maxlength = 1518 = 0x5EE
    proc_common_wait_some_cycles(mm_clk, 10);
    
    ---- . rx set primary SRC_MAC
    --proc_mem_mm_bus_wr(16#0802#,       c_mac0, mm_clk, mm_miso, mm_mosi);  -- src_mac(31 DOWNTO  0)
    --proc_mem_mm_bus_wr(16#0803#,       c_mac1, mm_clk, mm_miso, mm_mosi);  -- src_mac(47 DOWNTO 32) <-- primary SRC_MAC = 12-34-56-78-9A-BC
    --proc_mem_mm_bus_rd(16#0802#,               mm_clk, mm_miso, mm_mosi);  -- src_mac(31 DOWNTO  0)
    --proc_mem_mm_bus_rd(16#0803#,               mm_clk, mm_miso, mm_mosi);  -- src_mac(47 DOWNTO 32)
    --proc_common_wait_some_cycles(mm_clk, 10);
    --
    ---- . tx insert frame SRC_MAC
    --proc_mem_mm_bus_wr(16#1200#,            1, mm_clk, mm_miso, mm_mosi);  -- RW, tx_addrins_control = 1 to overwrite frame SRC_MAC with tx_addrins_macaddr0,1
    --proc_mem_mm_bus_wr(16#1201#,       c_mac0, mm_clk, mm_miso, mm_mosi);  -- src_mac(31 DOWNTO  0)
    --proc_mem_mm_bus_wr(16#1202#,       c_mac1, mm_clk, mm_miso, mm_mosi);  -- src_mac(47 DOWNTO 32) <-- frame SRC_MAC = 12-34-56-78-9A-BC
    --proc_common_wait_some_cycles(mm_clk, 10);
    
    WAIT UNTIL rising_edge(mm_clk);
  END proc_tech_mac_10g_setup_stratixiv;

  PROCEDURE proc_tech_mac_10g_setup_arria10(CONSTANT src_mac             : IN  STD_LOGIC_VECTOR(c_network_eth_mac_slv'RANGE);
                                            SIGNAL   mm_clk              : IN  STD_LOGIC;
                                            SIGNAL   mm_miso             : IN  t_mem_miso;
                                            SIGNAL   mm_mosi             : OUT t_mem_mosi) IS
  BEGIN
    proc_tech_mac_10g_setup_stratixiv(src_mac, mm_clk, mm_miso, mm_mosi);
  END proc_tech_mac_10g_setup_arria10;
  
  -- Transmit user packet
  -- . Use word aligned payload data, so with padding inserted before the 14 byte header
  -- . Packets can be send immediately after eachother so new sop directly after last eop
  -- . The word rate is controlled by respecting ready from the MAC
  PROCEDURE proc_tech_mac_10g_tx_packet(CONSTANT total_header    : IN  t_network_total_header;
                                        CONSTANT data_len        : IN  NATURAL;  -- in symbols = octets = bytes
                                        CONSTANT c_data_type     : IN  NATURAL;  -- c_tb_tech_mac_10g_data_type_*
                                        CONSTANT c_ready_latency : IN  NATURAL;  -- 0, 1 are supported by proc_dp_stream_ready_latency()
                                        CONSTANT c_nof_not_valid : IN  NATURAL;  -- when > 0 then pull tx valid low for c_nof_not_valid beats during tx
                                        SIGNAL   ff_clk          : IN  STD_LOGIC;
                                        SIGNAL   ff_en           : IN  STD_LOGIC;  -- similar purpose as c_nof_not_valid, but not used so pass on signal '1'
                                        SIGNAL   ff_src_in       : IN  t_dp_siso;
                                        SIGNAL   ff_src_out      : OUT t_dp_sosi) IS
    CONSTANT c_arp_words_arr  : t_network_total_header_64b_arr := func_network_total_header_construct_arp( total_header.eth, total_header.arp);
    CONSTANT c_icmp_words_arr : t_network_total_header_64b_arr := func_network_total_header_construct_icmp(total_header.eth, total_header.ip, total_header.icmp);
    CONSTANT c_udp_words_arr  : t_network_total_header_64b_arr := func_network_total_header_construct_udp( total_header.eth, total_header.ip, total_header.udp);
    
    CONSTANT c_mod            : NATURAL := data_len MOD c_tech_mac_10g_symbols_per_beat;
    CONSTANT c_nof_data_beats : NATURAL := data_len   / c_tech_mac_10g_symbols_per_beat + sel_a_b(c_mod, 1, 0);
    CONSTANT c_empty          : NATURAL := sel_a_b(c_mod, c_tech_mac_10g_symbols_per_beat - c_mod, 0);
    VARIABLE v_sym            : UNSIGNED(c_tech_mac_10g_symbol_w-1 DOWNTO 0) := (OTHERS=>'0');
    VARIABLE v_num            : UNSIGNED(c_tech_mac_10g_data_w-1 DOWNTO 0) := (OTHERS=>'0');
    VARIABLE v_hdr_words_arr  : t_network_total_header_64b_arr;
  BEGIN
    -- Select header
    CASE c_data_type IS
      WHEN c_tb_tech_mac_10g_data_type_arp  => v_hdr_words_arr := c_arp_words_arr;
      WHEN c_tb_tech_mac_10g_data_type_ping => v_hdr_words_arr := c_icmp_words_arr;
      WHEN OTHERS                           => v_hdr_words_arr := c_udp_words_arr;   -- default via UDP
    END CASE;
  
    ff_src_out.empty <= TO_DP_EMPTY(0);
    IF c_data_type=c_tb_tech_mac_10g_data_type_arp THEN
    
      ----------------------------------------------------------------------------
      -- Header only frame
      ----------------------------------------------------------------------------
      -- . sop
      ff_src_out.data <= RESIZE_DP_DATA(v_hdr_words_arr(0));
      proc_dp_stream_ready_latency(c_ready_latency, ff_clk, ff_src_in.ready, ff_en, '0', '1', '1', '0', ff_src_out.sync, ff_src_out.valid, ff_src_out.sop, ff_src_out.eop);
      ff_src_out.data <= RESIZE_DP_DATA(v_hdr_words_arr(1));  -- prepare data before loop, so proc_dp_stream_ready_latency can be called at start of the loops
      FOR I IN 2 TO c_nof_hdr_beats-2 LOOP
        proc_dp_stream_ready_latency(c_ready_latency, ff_clk, ff_src_in.ready, ff_en, '0', '1', '0', '0', ff_src_out.sync, ff_src_out.valid, ff_src_out.sop, ff_src_out.eop);
        ff_src_out.data <= RESIZE_DP_DATA(v_hdr_words_arr(I));
      END LOOP;
      proc_dp_stream_ready_latency(c_ready_latency, ff_clk, ff_src_in.ready, ff_en, '0', '1', '0', '0', ff_src_out.sync, ff_src_out.valid, ff_src_out.sop, ff_src_out.eop);
      -- . eop
      ff_src_out.data <= RESIZE_DP_DATA(v_hdr_words_arr(c_nof_hdr_beats-1));
      proc_dp_stream_ready_latency(c_ready_latency, ff_clk, ff_src_in.ready, ff_en, '0', '1', '0', '1', ff_src_out.sync, ff_src_out.valid, ff_src_out.sop, ff_src_out.eop);
    
    ELSE
    
      ----------------------------------------------------------------------------
      -- Header for frame with payload
      ----------------------------------------------------------------------------
    
      -- Header
      -- . sop
      ff_src_out.data <= RESIZE_DP_DATA(v_hdr_words_arr(0));
      proc_dp_stream_ready_latency(c_ready_latency, ff_clk, ff_src_in.ready, ff_en, '0', '1', '1', '0', ff_src_out.sync, ff_src_out.valid, ff_src_out.sop, ff_src_out.eop);
      ff_src_out.data <= RESIZE_DP_DATA(v_hdr_words_arr(1));  -- prepare data before loop, so proc_dp_stream_ready_latency can be called at start of the loops
      FOR I IN 2 TO c_nof_hdr_beats-1 LOOP
        proc_dp_stream_ready_latency(c_ready_latency, ff_clk, ff_src_in.ready, ff_en, '0', '1', '0', '0', ff_src_out.sync, ff_src_out.valid, ff_src_out.sop, ff_src_out.eop);
        ff_src_out.data <= RESIZE_DP_DATA(v_hdr_words_arr(I));
      END LOOP;

      -- Payload
      FOR I IN 0 TO c_nof_data_beats-1 LOOP
        proc_dp_stream_ready_latency(c_ready_latency, ff_clk, ff_src_in.ready, ff_en, '0', '1', '0', '0', ff_src_out.sync, ff_src_out.valid, ff_src_out.sop, ff_src_out.eop);
        CASE c_data_type IS
          WHEN c_tb_tech_mac_10g_data_type_counter =>
            -- data : X"00000000_00000001", X"00000000_00000002", X"00000000_00000003", etc
            v_num := v_num + 1;
            ff_src_out.data <= RESIZE_DP_DATA(STD_LOGIC_VECTOR(v_num));
          WHEN OTHERS =>
            -- data : X"01020304_05060708", X"090A0B0C_0D0E0F10", X"11121314_15161718", etc
            FOR J IN c_tech_mac_10g_symbols_per_beat-1 DOWNTO 0 LOOP
              v_sym := v_sym + 1;
              ff_src_out.data((J+1)*c_tech_mac_10g_symbol_w-1 DOWNTO J*c_tech_mac_10g_symbol_w) <= STD_LOGIC_VECTOR(v_sym);
            END LOOP;
        END CASE;
        -- tb : pull valid low for some time during the middle of the payload
        IF c_nof_not_valid > 0 AND I=c_nof_data_beats/2 THEN
          ff_src_out.valid <= '0';
          FOR I IN 0 TO c_nof_not_valid LOOP WAIT UNTIL rising_edge(ff_clk); END LOOP;
          ff_src_out.valid <= '1';
        END IF;
      END LOOP;
      
      -- Last data
      IF c_empty > 0 THEN
        -- Overwrite empty data
        ff_src_out.empty <= TO_DP_EMPTY(c_empty);
        FOR J IN c_empty-1 DOWNTO 0 LOOP
          ff_src_out.data((J+1)*c_tech_mac_10g_symbol_w-1 DOWNTO J*c_tech_mac_10g_symbol_w) <= (OTHERS=>'0');
        END LOOP;
      END IF;
      -- . eop
      proc_dp_stream_ready_latency(c_ready_latency, ff_clk, ff_src_in.ready, ff_en, '0', '1', '0', '1', ff_src_out.sync, ff_src_out.valid, ff_src_out.sop, ff_src_out.eop);
      
    END IF;
    
    -- Initialize for next tx packet
    ff_src_out.data  <= TO_DP_DATA(0);
    ff_src_out.valid <= '0';
    ff_src_out.eop   <= '0';
    ff_src_out.empty <= TO_DP_EMPTY(0);
  END proc_tech_mac_10g_tx_packet;

  
  -- Receive packet
  -- . Use word aligned payload data, so with padding inserted before the 14 byte header
  -- . Packets can be always be received, assume the user application is always ready
  -- . The CRC32 is also passed on to the user at eop.
  -- . Note that when empty/=0 then the CRC32 is not word aligned, so therefore use prev_data to be able
  --   to handle part of last data word in case empty/=0 at eop
  PROCEDURE proc_tech_mac_10g_rx_packet(CONSTANT total_header : IN  t_network_total_header;
                                        CONSTANT c_data_type  : IN  NATURAL;  -- c_tb_tech_mac_10g_data_type_*
                                        SIGNAL   ff_clk       : IN  STD_LOGIC;
                                        SIGNAL   ff_snk_in    : IN  t_dp_sosi;
                                        SIGNAL   ff_snk_out   : OUT t_dp_siso) IS
    CONSTANT c_eth_header     : t_network_eth_header := total_header.eth;
    CONSTANT c_arp_words_arr  : t_network_total_header_64b_arr := func_network_total_header_construct_arp( total_header.eth, total_header.arp);
    CONSTANT c_icmp_words_arr : t_network_total_header_64b_arr := func_network_total_header_construct_icmp(total_header.eth, total_header.ip, total_header.icmp);
    CONSTANT c_udp_words_arr  : t_network_total_header_64b_arr := func_network_total_header_construct_udp( total_header.eth, total_header.ip, total_header.udp);
    VARIABLE v_sym            : UNSIGNED(c_tech_mac_10g_symbol_w-1 DOWNTO 0) := (OTHERS=>'0');
    VARIABLE v_num            : UNSIGNED(c_tech_mac_10g_data_w-1 DOWNTO 0) := (OTHERS=>'0');
    VARIABLE v_empty          : NATURAL;
    VARIABLE v_first          : BOOLEAN := TRUE;
    VARIABLE v_data           : STD_LOGIC_VECTOR(c_tech_mac_10g_data_w-1 DOWNTO 0);
    VARIABLE v_prev_data      : STD_LOGIC_VECTOR(c_tech_mac_10g_data_w-1 DOWNTO 0);
    VARIABLE v_hdr_string     : STRING(1 TO 11);
    VARIABLE v_hdr_words_arr  : t_network_total_header_64b_arr;
  BEGIN
    -- Select header
    CASE c_data_type IS
      WHEN c_tb_tech_mac_10g_data_type_arp  => v_hdr_string := "ETH/ARP    "; v_hdr_words_arr := c_arp_words_arr;
      WHEN c_tb_tech_mac_10g_data_type_ping => v_hdr_string := "ETH/IP/ICMP"; v_hdr_words_arr := c_icmp_words_arr;
      WHEN OTHERS                           => v_hdr_string := "ETH/IP/UDP "; v_hdr_words_arr := c_udp_words_arr;   -- default via UDP
    END CASE;
    
    -- Keep ff_rx_snk_out.ready='1' and ff_rx_snk_out.xon='1' all the time
    ff_snk_out <= c_dp_siso_rdy;
    
    ----------------------------------------------------------------------------
    -- Verify ETH header part
    ----------------------------------------------------------------------------
    -- . wait for sop
    proc_dp_stream_valid_sop(ff_clk, ff_snk_in.valid, ff_snk_in.sop);
    ASSERT ff_snk_in.data(63 DOWNTO 16) = X"00000000_0000"                   REPORT "RX: Wrong ETH alignment padding not zero" SEVERITY ERROR;
    ASSERT ff_snk_in.data(15 DOWNTO  0) = c_eth_header.dst_mac(47 DOWNTO 32) REPORT "RX: Wrong ETH dst_mac_addr(47 DOWNTO 32)" SEVERITY ERROR;
    proc_dp_stream_valid(ff_clk, ff_snk_in.valid);
    ASSERT ff_snk_in.data(63 DOWNTO 32) = c_eth_header.dst_mac(31 DOWNTO 0)  REPORT "RX: Wrong ETH dst_mac_addr(31 DOWNTO 0)" SEVERITY ERROR;
    ASSERT ff_snk_in.data(31 DOWNTO  0) = c_eth_header.src_mac(47 DOWNTO 16) REPORT "RX: Wrong ETH src_mac_addr(47 DOWNTO 16)" SEVERITY ERROR;
    proc_dp_stream_valid(ff_clk, ff_snk_in.valid);
    ASSERT ff_snk_in.data(63 DOWNTO 48) = c_eth_header.src_mac(15 DOWNTO 0)  REPORT "RX: Wrong ETH src_mac_addr(15 DOWNTO 0)" SEVERITY ERROR;
    ASSERT ff_snk_in.data(47 DOWNTO 32) = c_eth_header.eth_type              REPORT "RX: Wrong ETH ethertype" SEVERITY ERROR;
    
    -- . verify remaining half word
    ASSERT ff_snk_in.data(31 DOWNTO 0) = v_hdr_words_arr(c_nof_eth_beats-1)(31 DOWNTO 0) REPORT "RX: Wrong " & v_hdr_string & " response half word" SEVERITY ERROR;
    
    IF c_data_type=c_tb_tech_mac_10g_data_type_arp THEN
    
      ----------------------------------------------------------------------------
      -- Verify header only frame
      ----------------------------------------------------------------------------

      FOR I IN c_nof_eth_beats TO c_nof_hdr_beats-1 LOOP
        proc_dp_stream_valid(ff_clk, ff_snk_in.valid);
        ASSERT ff_snk_in.data(63 DOWNTO 0) = v_hdr_words_arr(I) REPORT "RX: Wrong " & v_hdr_string & " response word" SEVERITY ERROR;
      END LOOP;
      -- . continue to eop
      WHILE ff_snk_in.eop /= '1' LOOP
        proc_dp_stream_valid(ff_clk, ff_snk_in.valid);
      END LOOP;
      
    ELSE
    
      ----------------------------------------------------------------------------
      -- Verify header and payload
      ----------------------------------------------------------------------------
    
      -- Header
      FOR I IN c_nof_eth_beats TO c_nof_hdr_beats-1 LOOP
        proc_dp_stream_valid(ff_clk, ff_snk_in.valid);
        IF I/=c_network_total_header_64b_ip_header_checksum_wi THEN  -- do not verify tx ip header checksum
          ASSERT ff_snk_in.data(63 DOWNTO 0) = v_hdr_words_arr(I) REPORT "RX: Wrong " & v_hdr_string & " response word" SEVERITY ERROR;
        END IF;
      END LOOP;
      
      -- Payload
      -- . continue to eop
      v_first := TRUE;
      proc_dp_stream_valid(ff_clk, ff_snk_in.valid);
      WHILE ff_snk_in.eop /= '1' LOOP
        v_prev_data := v_data;
        v_data      := ff_snk_in.data(63 DOWNTO 0);
        IF v_first = FALSE THEN
          CASE c_data_type IS
            WHEN c_tb_tech_mac_10g_data_type_counter =>
              -- data : X"00000000_00000001", X"00000000_00000002", X"00000000_00000003", etc
              v_num := v_num + 1;
              IF UNSIGNED(v_prev_data)/=0 THEN   -- do not verify zero padding
                ASSERT UNSIGNED(v_prev_data) = v_num REPORT "RX: Wrong data word" SEVERITY ERROR;
              END IF;
            WHEN OTHERS =>
              -- data : X"01020304_05060708", X"090A0B0C_0D0E0F10", X"11121314_15161718", etc
              FOR J IN c_tech_mac_10g_symbols_per_beat-1 DOWNTO 0 LOOP
                v_sym := v_sym + 1;
                IF UNSIGNED(v_prev_data((J+1)*c_tech_mac_10g_symbol_w-1 DOWNTO J*c_tech_mac_10g_symbol_w))/=0 THEN   -- do not verify zero padding
                  ASSERT UNSIGNED(v_prev_data((J+1)*c_tech_mac_10g_symbol_w-1 DOWNTO J*c_tech_mac_10g_symbol_w)) = v_sym REPORT "RX: Wrong data symbol" SEVERITY ERROR;
                END IF;
              END LOOP;
          END CASE;
        END IF;
        v_first := FALSE;
        proc_dp_stream_valid(ff_clk, ff_snk_in.valid);
      END LOOP;
      
      -- Verify last data and CRC32 if empty<4 else the last word is only the CRC32
      v_prev_data := v_data;
      v_data      := ff_snk_in.data(63 DOWNTO 0);
      v_empty     := TO_INTEGER(UNSIGNED(ff_snk_in.empty(c_tech_mac_10g_empty_w-1 DOWNTO 0)));
      IF v_empty < 4 THEN
        FOR J IN v_empty-1 DOWNTO 0 LOOP
          v_prev_data((J+1)*c_tech_mac_10g_symbol_w-1 DOWNTO J*c_tech_mac_10g_symbol_w) := (OTHERS=>'0');
        END LOOP;
        CASE c_data_type IS
          WHEN c_tb_tech_mac_10g_data_type_counter =>
            -- data : X"00000000_00000001", X"00000000_00000002", X"00000000_00000003", etc
            v_num := v_num + 1;
            FOR J IN v_empty-1 DOWNTO 0 LOOP
              v_num((J+1)*c_tech_mac_10g_symbol_w-1 DOWNTO J*c_tech_mac_10g_symbol_w) := (OTHERS=>'0');  -- force CRC32 symbols in last data word to 0
            END LOOP;
            IF UNSIGNED(v_prev_data)/=0 THEN   -- do not verify zero padding
              ASSERT UNSIGNED(v_prev_data) = v_num REPORT "RX: Wrong empty data word" SEVERITY ERROR;
            END IF;
          WHEN OTHERS =>
            -- data : X"01020304_05060708", X"090A0B0C_0D0E0F10", X"11121314_15161718", etc
            FOR J IN c_tech_mac_10g_symbols_per_beat-1 DOWNTO v_empty LOOP  -- ignore CRC32 symbols in last data word
              v_sym := v_sym + 1;
              IF UNSIGNED(v_prev_data((J+1)*c_tech_mac_10g_symbol_w-1 DOWNTO J*c_tech_mac_10g_symbol_w))/=0 THEN   -- do not verify zero padding
                ASSERT UNSIGNED(v_prev_data((J+1)*c_tech_mac_10g_symbol_w-1 DOWNTO J*c_tech_mac_10g_symbol_w)) = v_sym REPORT "RX: Wrong empty data symbol" SEVERITY ERROR;
              END IF;
            END LOOP;
        END CASE;
      END IF;
    END IF;
    
    -- No verify on CRC32 word
  END proc_tech_mac_10g_rx_packet;

END tb_tech_mac_10g_pkg;
