--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------


-- Purpose:
-- Description:
--
--   If g_pre_header_padding is TRUE then remove pre header padding from ST tx
--   and insert it in ST rx, because the 10G MAC IP does not use pre header
--   padding. When g_pre_header_padding is FALSE then the ST frame is directly
--   passed on via the dp_latency_adapters. The dp_latency_adapters are needed
--   because the ST interface uses RL=1 whereas e.g. the Altera mac_10g IP uses
--   RL=0.
--
--                                            tx_remove_snk      tx_mac_snk
--              __________                     .  ______________  .  __________________
--              |        |                     .  |            |  .  |                |
--              | dp_pad |                     .  | dp_latency |  .  |                |
--    tx_snk -->| remove |----------------------->| adapter    |---->|                |
--              |        |                        | RL=1 --> 0 |     |                |
--              |________|                        |____________|     |                |
--                                                                   |tech_mac_10g    |
--              __________     ______________     ______________     |ip_<device_name>|---XGMII
--              |        |     |            |     |            |     |                |
--              | dp_pad |     | dp_latency |     | dp_latency |     |                |
--    rx_src <--| insert |<----| fifo       |<----| adapter    |<----|                |
--              |        |  .  |            |  .  | RL=1 --> 0 |  .  |                |
--              |________|  .  |____________|  .  |____________|  .  |________________|
--                          .                  .                  .           |
--                          x_fifo_src        rx_mac_src_rl1     rx_mac_src   |
--                                                                            |
--                                                                           csr
--
-- Remarks:
-- . The dp_latency_fifo is necessary because the rx_mac_src_in.ready must remain
--   active during a reception, otherwise the rx frame gets truncated due to
--   back pressure by dp_pad_insert.
--

LIBRARY IEEE, common_lib, dp_lib, technology_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE technology_lib.technology_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_interface_layers_pkg.ALL;
USE common_lib.common_network_total_header_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE work.tech_mac_10g_component_pkg.ALL;

ENTITY tech_mac_10g IS
  GENERIC (
    g_technology          : NATURAL := c_tech_select_default;
    g_pre_header_padding  : BOOLEAN := FALSE
  );
  PORT (
    -- MM
    mm_clk           : IN  STD_LOGIC;
    mm_rst           : IN  STD_LOGIC;
    csr_mosi         : IN  t_mem_mosi;  -- CSR = control status register
    csr_miso         : OUT t_mem_miso;

    -- ST
    tx_clk_2x        : IN  STD_LOGIC := '0';   -- 312.5 MHz
    tx_clk           : IN  STD_LOGIC;   -- 156.25 MHz local reference
    tx_rst           : IN  STD_LOGIC;
    tx_snk_in        : IN  t_dp_sosi;   -- 64 bit data
    tx_snk_out       : OUT t_dp_siso; 
    
    rx_clk_2x        : IN  STD_LOGIC := '0';   -- 312.5 MHz
    rx_clk           : IN  STD_LOGIC;   -- 156.25 MHz from rx phy
    rx_rst           : IN  STD_LOGIC;
    rx_src_out       : OUT t_dp_sosi;   -- 64 bit data
    rx_src_in        : IN  t_dp_siso; 
    
    -- XGMII
    xgmii_tx_data    : OUT STD_LOGIC_VECTOR(c_xgmii_w-1 DOWNTO 0);  -- 72 bit
    xgmii_rx_data    : IN  STD_LOGIC_VECTOR(c_xgmii_w-1 DOWNTO 0)   -- 72 bit
  );
END tech_mac_10g;


ARCHITECTURE str OF tech_mac_10g IS

  -- Adapt ST ready latency 1 to IP ready latency
  CONSTANT c_ip_tx_ready_latency : NATURAL := sel_a_b(g_technology=c_tech_stratixiv OR g_technology=c_tech_arria10, 0, 1);
  CONSTANT c_ip_rx_ready_latency : NATURAL := sel_a_b(g_technology=c_tech_stratixiv OR g_technology=c_tech_arria10, 0, 1);

  SIGNAL tx_mac_snk_in_data  : STD_LOGIC_VECTOR(c_tech_mac_10g_data_w-1 DOWNTO 0);  -- 64 bit
  SIGNAL tx_mac_snk_in       : t_dp_sosi;
  SIGNAL tx_mac_snk_out      : t_dp_siso;
  SIGNAL tx_remove_snk_in    : t_dp_sosi;
  SIGNAL tx_remove_snk_out   : t_dp_siso;
  
  SIGNAL rx_mac_src_out_data : STD_LOGIC_VECTOR(c_tech_mac_10g_data_w-1 DOWNTO 0);  -- 64 bit
  SIGNAL rx_mac_src_out      : t_dp_sosi := c_dp_sosi_rst;  -- force unused bits to '0' to ease monitoring in wave window
  SIGNAL rx_mac_src_in       : t_dp_siso;
  SIGNAL rx_mac_src_out_rl1  : t_dp_sosi;
  SIGNAL rx_mac_src_in_rl1   : t_dp_siso;
  SIGNAL rx_fifo_src_out     : t_dp_sosi;
  SIGNAL rx_fifo_src_in      : t_dp_siso;
  SIGNAL rx_insert_src_out   : t_dp_sosi;
  SIGNAL rx_insert_src_in    : t_dp_siso;
              
  CONSTANT c_fifo_size       : NATURAL := 2;  -- latency to insert pre header padding, can take ready low for 2 cycles 
  
  SIGNAL fifo_usedw          : STD_LOGIC_VECTOR(ceil_log2(3+c_fifo_size)-1 DOWNTO 0);  -- +3 to match dp_latency_fifo usedw width
  SIGNAL fifo_ful            : STD_LOGIC;
  SIGNAL fifo_emp            : STD_LOGIC;
  
BEGIN
 
  gen_ip_stratixiv : IF g_technology=c_tech_stratixiv GENERATE
    u0 : ENTITY work.tech_mac_10g_stratixiv
    PORT MAP (mm_clk, mm_rst, csr_mosi, csr_miso,
              tx_clk, tx_rst, tx_mac_snk_in, tx_mac_snk_out,
              rx_clk, rx_rst, rx_mac_src_out, rx_mac_src_in,
              xgmii_tx_data, xgmii_rx_data);
  END GENERATE;
  
  gen_ip_arria10 : IF g_technology=c_tech_arria10 GENERATE
    u0 : ENTITY work.tech_mac_10g_arria10
    PORT MAP (mm_clk, mm_rst, csr_mosi, csr_miso,
              tx_clk_2x, tx_clk, tx_rst, tx_mac_snk_in, tx_mac_snk_out,
              rx_clk_2x, rx_clk, rx_rst, rx_mac_src_out, rx_mac_src_in,
              xgmii_tx_data, xgmii_rx_data);  
  END GENERATE;
  
  
  -----------------------------------------------------------------------------
  -- Debug signals to ease monitoring in wave window  
  -----------------------------------------------------------------------------
  
  tx_mac_snk_in_data  <= tx_mac_snk_in.data(c_tech_mac_10g_data_w-1 DOWNTO 0);
  rx_mac_src_out_data <= rx_mac_src_out.data(c_tech_mac_10g_data_w-1 DOWNTO 0);
  
  
  -----------------------------------------------------------------------------
  -- Tx
  -----------------------------------------------------------------------------

  -- If g_pre_header_padding is TRUE then tx remove pre header padding else bypass
  gen_tx_remove : IF g_pre_header_padding=TRUE GENERATE
    u_tx_dp_pad_remove : ENTITY dp_lib.dp_pad_remove
    GENERIC MAP (
      g_data_w          => c_tech_mac_10g_data_w,
      g_symbol_w        => c_tech_mac_10g_symbol_w,
      g_nof_padding     => c_network_total_header_64b_align_len
    )
    PORT MAP (
      rst     => tx_rst,
      clk     => tx_clk,
      snk_out => tx_snk_out,
      snk_in  => tx_snk_in,
      src_in  => tx_remove_snk_out,
      src_out => tx_remove_snk_in
    );
  END GENERATE;
  
  no_tx_remove : IF g_pre_header_padding=FALSE GENERATE
    tx_snk_out       <= tx_remove_snk_out;
    tx_remove_snk_in <= tx_snk_in;
  END GENERATE;
  
 
  u_tx_dp_latency_adapter : ENTITY dp_lib.dp_latency_adapter
  GENERIC MAP (
    g_in_latency  => 1,
    g_out_latency => c_ip_tx_ready_latency
  )
  PORT MAP (
    rst       => tx_rst,
    clk       => tx_clk,
    snk_out   => tx_remove_snk_out,
    snk_in    => tx_remove_snk_in,
    src_in    => tx_mac_snk_out,
    src_out   => tx_mac_snk_in
  );
  
  
  -----------------------------------------------------------------------------
  -- Rx
  -----------------------------------------------------------------------------
  
  u_rx_dp_latency_adapter : ENTITY dp_lib.dp_latency_adapter
  GENERIC MAP (
    g_in_latency  => c_ip_rx_ready_latency,
    g_out_latency => 1
  )
  PORT MAP (
    rst       => rx_rst,
    clk       => rx_clk,
    snk_out   => rx_mac_src_in,
    snk_in    => rx_mac_src_out,
    src_in    => rx_mac_src_in_rl1,
    src_out   => rx_mac_src_out_rl1
  );
    
  -- If g_pre_header_padding is TRUE then rx insert pre header padding to data word align the payload else bypass
  gen_rx_insert : IF g_pre_header_padding=TRUE GENERATE
    -- Need FIFO to avoid that with dp_pad_insert flow control stops the rx MAC
    u_rx_dp_latency_fifo : ENTITY dp_lib.dp_latency_fifo
    GENERIC MAP (
      g_fifo_size  => c_fifo_size
    )
    PORT MAP (
      rst       => rx_rst,
      clk       => rx_clk,
      usedw     => fifo_usedw,
      wr_ful    => fifo_ful,
      rd_emp    => fifo_emp,
      snk_out   => rx_mac_src_in_rl1,
      snk_in    => rx_mac_src_out_rl1,
      src_in    => rx_fifo_src_in,
      src_out   => rx_fifo_src_out
    );
  
    u_rx_dp_pad_insert : ENTITY dp_lib.dp_pad_insert
    GENERIC MAP (
      g_data_w          => c_tech_mac_10g_data_w,
      g_symbol_w        => c_tech_mac_10g_symbol_w,
      g_nof_padding     => c_network_total_header_64b_align_len
    )
    PORT MAP (
      rst     => rx_rst,
      clk     => rx_clk,
      snk_out => rx_fifo_src_in,
      snk_in  => rx_fifo_src_out,  
      src_in  => rx_insert_src_in,
      src_out => rx_insert_src_out
    );
    
    rx_insert_src_in <= rx_src_in;
    rx_src_out       <= rx_insert_src_out;
  END GENERATE;
  
  no_rx_insert : IF g_pre_header_padding=FALSE GENERATE
    rx_mac_src_in_rl1 <= rx_src_in;
    rx_src_out        <= rx_mac_src_out_rl1;
  END GENERATE;
   
END str;

