--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- Declare IP libraries to ensure default binding in simulation. The IP library clause is ignored by synthesis.
LIBRARY ip_stratixiv_transceiver_lib, ip_stratixiv_phy_xaui_lib;

LIBRARY IEEE, technology_lib, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_interface_layers_pkg.ALL;
USE technology_lib.technology_pkg.ALL;
USE work.tech_xaui_component_pkg.ALL;

ENTITY tech_xaui_stratixiv IS
  GENERIC (
    g_nof_xaui        : NATURAL := 1 -- Up to 3 (hard XAUI only) supported
  );                      
  PORT (   
    -- Transceiver PLL reference clock   
    tr_clk                  : IN  STD_LOGIC;

    mm_clk                  : IN  STD_LOGIC;
    mm_rst                  : IN  STD_LOGIC;

    -- Calibration & reconfig clock
    cal_rec_clk             : IN  STD_LOGIC;
 
    tx_clk_arr              : IN  STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
    rx_clk_arr              : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
   
    crc_rx_ready_arr        : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0); --crc = synchronous to Cal_Rec_Clk
    crc_tx_ready_arr        : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);

    a_rx_channelaligned_arr : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);

    xgmii_tx_dc_arr         : IN  t_xgmii_dc_arr(g_nof_xaui-1 DOWNTO 0);
    xgmii_rx_dc_arr         : OUT t_xgmii_dc_arr(g_nof_xaui-1 DOWNTO 0);

    xaui_rx_arr             : IN  t_xaui_arr(g_nof_xaui-1 DOWNTO 0);
    xaui_tx_arr             : OUT t_xaui_arr(g_nof_xaui-1 DOWNTO 0);

    xaui_mosi               : IN  t_mem_mosi := c_mem_mosi_rst;
    xaui_miso               : OUT t_mem_miso
  );
END tech_xaui_stratixiv;


ARCHITECTURE str OF tech_xaui_stratixiv IS

  CONSTANT c_reconf_togxb_bus_w         : NATURAL := 4; 
  CONSTANT c_reconf_fromgxb_bus_w       : NATURAL := 17;
  CONSTANT c_reconf_nof_gx              : NATURAL := 4;                         -- 1, 2, or 3 hard XAUI with 4, 8 or 12 transceivers all require only gxb_reconfig for 4 transceivers
  CONSTANT c_reconf_soft_nof_gx         : NATURAL := 16;                        -- g_nof_xaui=4 soft XAUI, with 16 transceivers requires gxb_reconfig as for 16 transceivers
  CONSTANT c_reconf_soft_fromgxb_bus_w  : NATURAL := 4*c_reconf_fromgxb_bus_w;

  CONSTANT c_xaui_mosi_addr_w           : NATURAL := tech_xaui_mosi_addr_w(c_tech_stratixiv);

  SIGNAL reconfig_togxb            : STD_LOGIC_VECTOR(c_reconf_togxb_bus_w-1 DOWNTO 0);
  SIGNAL reconfig_fromgxb          : STD_LOGIC_VECTOR(c_reconf_fromgxb_bus_w-1 DOWNTO 0); 

  SIGNAL reconfig_soft_togxb       : STD_LOGIC_VECTOR(c_reconf_togxb_bus_w-1 DOWNTO 0);
  SIGNAL reconfig_soft_fromgxb     : STD_LOGIC_VECTOR(c_reconf_soft_fromgxb_bus_w-1 DOWNTO 0);

  SIGNAL reconfig_0_togxb          : STD_LOGIC_VECTOR(c_reconf_togxb_bus_w-1 DOWNTO 0);
  SIGNAL reconfig_0_fromgxb        : STD_LOGIC_VECTOR(c_reconf_fromgxb_bus_w-1 DOWNTO 0);
  SIGNAL reconfig_1_togxb          : STD_LOGIC_VECTOR(c_reconf_togxb_bus_w-1 DOWNTO 0);
  SIGNAL reconfig_1_fromgxb        : STD_LOGIC_VECTOR(c_reconf_fromgxb_bus_w-1 DOWNTO 0);
  SIGNAL reconfig_2_togxb          : STD_LOGIC_VECTOR(c_reconf_togxb_bus_w-1 DOWNTO 0);
  SIGNAL reconfig_2_fromgxb        : STD_LOGIC_VECTOR(c_reconf_fromgxb_bus_w-1 DOWNTO 0);

  SIGNAL xaui_mosi_arr             : t_mem_mosi_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL xaui_miso_arr             : t_mem_miso_arr(g_nof_xaui-1 DOWNTO 0); 
 
BEGIN
 
  gen_nof_xaui : FOR i IN g_nof_xaui-1 DOWNTO 0 GENERATE

    gen_hard_xaui_0: IF i=0 GENERATE
      u_ip_phy_xaui : ip_stratixiv_phy_xaui_0
      PORT MAP (
      	pll_ref_clk          => tr_clk,
      	xgmii_tx_clk         => tx_clk_arr(i),
      	xgmii_rx_clk         => rx_clk_arr(i),
      	xgmii_rx_dc          => xgmii_rx_dc_arr(i),
      	xgmii_tx_dc          => xgmii_tx_dc_arr(i),
      	xaui_rx_serial_data  => xaui_rx_arr(i),
      	xaui_tx_serial_data  => xaui_tx_arr(i),
      	rx_ready             => crc_rx_ready_arr(i),
      	tx_ready             => crc_tx_ready_arr(i),
      	phy_mgmt_clk         => mm_clk,
      	phy_mgmt_clk_reset   => mm_rst,
      	phy_mgmt_address     => xaui_mosi_arr(i).address(c_xaui_mosi_addr_w-1 DOWNTO 0),
      	phy_mgmt_read        => xaui_mosi_arr(i).rd,
      	phy_mgmt_readdata    => xaui_miso_arr(i).rddata(c_word_w-1 DOWNTO 0),
      	phy_mgmt_write       => xaui_mosi_arr(i).wr,
      	phy_mgmt_writedata   => xaui_mosi_arr(i).wrdata(c_word_w-1 DOWNTO 0),
      	phy_mgmt_waitrequest => xaui_miso_arr(i).waitrequest,

        rx_channelaligned    => a_rx_channelaligned_arr(i),

        -- Reconfig block connections
        reconfig_to_xcvr     => reconfig_0_togxb,
        reconfig_from_xcvr   => reconfig_0_fromgxb
      );

      u_ip_gxb_reconfig : ip_stratixiv_gxb_reconfig_v111
      GENERIC MAP (
        g_nof_gx  => c_reconf_nof_gx
      )
      PORT MAP (
        reconfig_clk        => cal_rec_clk,
        reconfig_fromgxb    => reconfig_0_fromgxb,
        busy                => OPEN,
        reconfig_togxb      => reconfig_0_togxb
      );  
    END GENERATE;

    gen_hard_xaui_1: IF i=1 GENERATE
      u_ip_phy_xaui : ip_stratixiv_phy_xaui_1
      PORT MAP (
      	pll_ref_clk          => tr_clk,
      	xgmii_tx_clk         => tx_clk_arr(i),
      	xgmii_rx_clk         => rx_clk_arr(i),
      	xgmii_rx_dc          => xgmii_rx_dc_arr(i),
      	xgmii_tx_dc          => xgmii_tx_dc_arr(i),
      	xaui_rx_serial_data  => xaui_rx_arr(i),
      	xaui_tx_serial_data  => xaui_tx_arr(i),
      	rx_ready             => crc_rx_ready_arr(i),
      	tx_ready             => crc_tx_ready_arr(i),
      	phy_mgmt_clk         => mm_clk,
      	phy_mgmt_clk_reset   => mm_rst,
      	phy_mgmt_address     => xaui_mosi_arr(i).address(c_xaui_mosi_addr_w-1 DOWNTO 0),
      	phy_mgmt_read        => xaui_mosi_arr(i).rd,
      	phy_mgmt_readdata    => xaui_miso_arr(i).rddata(c_word_w-1 DOWNTO 0),
      	phy_mgmt_write       => xaui_mosi_arr(i).wr,
      	phy_mgmt_writedata   => xaui_mosi_arr(i).wrdata(c_word_w-1 DOWNTO 0),
      	phy_mgmt_waitrequest => xaui_miso_arr(i).waitrequest,

        rx_channelaligned    => a_rx_channelaligned_arr(i),

        -- Reconfig block connections
        reconfig_to_xcvr     => reconfig_1_togxb,
        reconfig_from_xcvr   => reconfig_1_fromgxb
      );

      u_ip_gxb_reconfig : ip_stratixiv_gxb_reconfig_v111
      GENERIC MAP (
        g_nof_gx  => c_reconf_nof_gx
      )
      PORT MAP (
        reconfig_clk        => cal_rec_clk,
        reconfig_fromgxb    => reconfig_1_fromgxb,
        busy                => OPEN,
        reconfig_togxb      => reconfig_1_togxb
      );  
    END GENERATE;

    gen_hard_xaui_2: IF i=2 GENERATE
      u_ip_phy_xaui : ip_stratixiv_phy_xaui_2
      PORT MAP (
      	pll_ref_clk          => tr_clk,
      	xgmii_tx_clk         => tx_clk_arr(i),
      	xgmii_rx_clk         => rx_clk_arr(i),
      	xgmii_rx_dc          => xgmii_rx_dc_arr(i),
      	xgmii_tx_dc          => xgmii_tx_dc_arr(i),
      	xaui_rx_serial_data  => xaui_rx_arr(i),
      	xaui_tx_serial_data  => xaui_tx_arr(i),
      	rx_ready             => crc_rx_ready_arr(i),
      	tx_ready             => crc_tx_ready_arr(i),
      	phy_mgmt_clk         => mm_clk,
      	phy_mgmt_clk_reset   => mm_rst,
      	phy_mgmt_address     => xaui_mosi_arr(i).address(c_xaui_mosi_addr_w-1 DOWNTO 0),
      	phy_mgmt_read        => xaui_mosi_arr(i).rd,
      	phy_mgmt_readdata    => xaui_miso_arr(i).rddata(c_word_w-1 DOWNTO 0),
      	phy_mgmt_write       => xaui_mosi_arr(i).wr,
      	phy_mgmt_writedata   => xaui_mosi_arr(i).wrdata(c_word_w-1 DOWNTO 0),
      	phy_mgmt_waitrequest => xaui_miso_arr(i).waitrequest,

        rx_channelaligned    => a_rx_channelaligned_arr(i),

        -- Reconfig block connections
        reconfig_to_xcvr     => reconfig_2_togxb,
        reconfig_from_xcvr   => reconfig_2_fromgxb
      );

      u_ip_gxb_reconfig : ip_stratixiv_gxb_reconfig_v111
      GENERIC MAP (
        g_nof_gx  => c_reconf_nof_gx
      )
      PORT MAP (
        reconfig_clk        => cal_rec_clk,
        reconfig_fromgxb    => reconfig_2_fromgxb,
        busy                => OPEN,
        reconfig_togxb      => reconfig_2_togxb
      );  
    END GENERATE;

    gen_soft_xaui: IF i=3 GENERATE -- NOTE: this 4th (soft) instance makes the Quartus fitter (11.1, no SP) fail, so is not supported.
      u_ip_phy_xaui : ip_stratixiv_phy_xaui_soft
      PORT MAP (
      	pll_ref_clk          => tr_clk,
      	xgmii_tx_clk         => tx_clk_arr(i),
      	xgmii_rx_clk         => rx_clk_arr(i),
      	xgmii_rx_dc          => xgmii_rx_dc_arr(i),
      	xgmii_tx_dc          => xgmii_tx_dc_arr(i),
      	xaui_rx_serial_data  => xaui_rx_arr(i),
      	xaui_tx_serial_data  => xaui_tx_arr(i),
      	rx_ready             => crc_rx_ready_arr(i),
      	tx_ready             => crc_tx_ready_arr(i),
      	phy_mgmt_clk         => mm_clk,
      	phy_mgmt_clk_reset   => mm_rst,
      	phy_mgmt_address     => xaui_mosi_arr(i).address(c_xaui_mosi_addr_w-1 DOWNTO 0),
      	phy_mgmt_read        => xaui_mosi_arr(i).rd,
      	phy_mgmt_readdata    => xaui_miso_arr(i).rddata(c_word_w-1 DOWNTO 0),
      	phy_mgmt_write       => xaui_mosi_arr(i).wr,
      	phy_mgmt_writedata   => xaui_mosi_arr(i).wrdata(c_word_w-1 DOWNTO 0),
      	phy_mgmt_waitrequest => xaui_miso_arr(i).waitrequest,

        rx_channelaligned    => a_rx_channelaligned_arr(i),

        -- Reconfig block connections
        reconfig_to_xcvr     => reconfig_soft_togxb,
        reconfig_from_xcvr   => reconfig_soft_fromgxb 
      );
      
      u_ip_gxb_reconfig : ip_stratixiv_gxb_reconfig_v111
      GENERIC MAP (
        g_nof_gx  => c_reconf_soft_nof_gx
      )
      PORT MAP (
        reconfig_clk        => cal_rec_clk,
        reconfig_fromgxb    => reconfig_soft_fromgxb,
        busy                => OPEN,
        reconfig_togxb      => reconfig_soft_togxb
      );  
    END GENERATE;

  END GENERATE;

  -----------------------------------------------------------------------------
  -- MM bus mux
  -----------------------------------------------------------------------------
  u_common_mem_mux : ENTITY common_lib.common_mem_mux
  GENERIC MAP (    
    g_nof_mosi    => g_nof_xaui,
    g_mult_addr_w => c_xaui_mosi_addr_w 
  )
  PORT MAP (
    mosi     => xaui_mosi,
    miso     => xaui_miso,
    mosi_arr => xaui_mosi_arr,
    miso_arr => xaui_miso_arr
  );
      
END str;

